from autocti.pipeline.phase.dataset import Result as DatasetResult


class Result(DatasetResult):
    @property
    def max_log_likelihood_full_fits(self):
        return self.analysis.fits_full_dataset_from_instance(
            instance=self.instance, hyper_noise_scale=True
        )

    @property
    def max_log_likelihood_full_fits_no_hyper_scaling(self):
        return self.analysis.fits_full_dataset_from_instance(
            instance=self.instance, hyper_noise_scale=False
        )

    @property
    def noise_scaling_maps_list_of_ci_regions(self):

        return list(
            map(
                lambda fit: fit.chi_squared_map_of_ci_regions,
                self.max_log_likelihood_full_fits_no_hyper_scaling,
            )
        )

    @property
    def noise_scaling_maps_list_of_parallel_trails(self):

        return list(
            map(
                lambda max_log_likelihood_full_fit: max_log_likelihood_full_fit.chi_squared_map_of_parallel_trails,
                self.max_log_likelihood_full_fits_no_hyper_scaling,
            )
        )

    @property
    def noise_scaling_maps_list_of_serial_trails(self):

        return list(
            map(
                lambda max_log_likelihood_full_fit: max_log_likelihood_full_fit.chi_squared_map_of_serial_trails,
                self.max_log_likelihood_full_fits_no_hyper_scaling,
            )
        )

    @property
    def noise_scaling_maps_list_of_serial_overscan_no_trails(self):

        return list(
            map(
                lambda max_log_likelihood_full_fit: max_log_likelihood_full_fit.chi_squared_map_of_serial_overscan_no_trails,
                self.max_log_likelihood_full_fits_no_hyper_scaling,
            )
        )



from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)

from autoarray.plot.mat_objs import Units
from autoarray.plot.mat_objs import Figure
from autoarray.plot.mat_objs import ColorMap
from autoarray.plot.mat_objs import ColorBar
from autoarray.plot.mat_objs import Ticks
from autoarray.plot.mat_objs import Labels
from autoarray.plot.mat_objs import Legend
from autoarray.plot.mat_objs import Output
from autoarray.plot.mat_objs import OriginScatterer
from autoarray.plot.mat_objs import Liner
from autoarray.plot.mat_objs import ParallelOverscanLiner
from autoarray.plot.mat_objs import SerialPrescanLiner
from autoarray.plot.mat_objs import SerialOverscanLiner

from autoarray.plot.plotters import Plotter
from autoarray.plot.plotters import SubPlotter
from autoarray.plot.plotters import Include

from autoarray.plot.plotters import plot_frame as Frame
from autoarray.plot.plotters import plot_line as Line

from autoarray.plot import imaging_plots as Imaging

from autocti.plot import ci_imaging_plots as CIImaging
from autocti.plot import ci_fit_plots as CIFit
from autocti.plot import ci_line_plots as CILine



import autocti as ac
import autofit as af
from test_autocti.integration.tests import runner

test_type = "features/total_density_range"
test_name = "parallel_x1__serial_x1"

test_path = "{}/../../../".format(os.path.dirname(os.path.realpath(__file__)))
output_path = test_path + "output/"
config_path = test_path + "config"
conf.instance = conf.Config(config_path=config_path, output_path=output_path)


clocker = ac.Clocker(parallel_express=2, serial_express=2)


def make_pipeline(name, folders, search=af.DynestyStatic()):
    class PhaseCIImaging(ac.PhaseCIImaging):
        def customize_priors(self, results):

            self.parallel_ccd.well_fill_alpha = 1.0
            self.parallel_ccd.well_fill_gamma = 0.0
            self.serial_ccd.well_fill_alpha = 1.0
            self.serial_ccd.well_fill_gamma = 0.0

    phase1 = ac.PhaseCIImaging(
        name="phase_1",
        folders=folders,
        search=search,
        parallel_traps=[af.PriorModel(ac.TrapInstantCapture)],
        parallel_ccd=parallel_ccd,
        serial_traps=[af.PriorModel(ac.TrapInstantCapture)],
        serial_ccd=serial_ccd,
        parallel_total_density_range=(0.1, 0.3),
        serial_total_density_range=(0.1, 0.3),
    )

    phase1.search.n_live_points = 60
    phase1.search.const_efficiency_mode = True
    phase1.search.facc = 0.2

    return ac.Pipeline(name, phase1)


if __name__ == "__main__":

    import sys

    runner.run(sys.modules[__name__], clocker=clocker)



import numpy as np
from autoconf import conf
from autocti.mask.mask import Mask2D


class SettingsCIMask:
    def __init__(
        self,
        parallel_front_edge_rows=None,
        parallel_trails_rows=None,
        serial_front_edge_columns=None,
        serial_trails_columns=None,
    ):

        self.parallel_front_edge_rows = parallel_front_edge_rows
        self.parallel_trails_rows = parallel_trails_rows
        self.serial_front_edge_columns = serial_front_edge_columns
        self.serial_trails_columns = serial_trails_columns

    @property
    def tag(self):
        return (
            f"{conf.instance['notation']['settings_tags']['ci_mask']['ci_mask']}["
            f"{self.parallel_front_edge_rows_tag}"
            f"{self.parallel_trails_rows_tag}"
            f"{self.serial_front_edge_columns_tag}"
            f"{self.serial_trails_columns_tag}]"
        )

    @property
    def parallel_front_edge_rows_tag(self):
        """Generate a parallel_front_edge_rows tag, to customize phase names based on the number of rows in the charge
        injection region at the front edge of the parallel clocking direction are masked during the fit,

        This changes the phase settings folder as follows:

        parallel_front_edge_rows = None -> settings
        parallel_front_edge_rows = (0, 10) -> settings__parallel_front_edge_rows_(0,10)
        parallel_front_edge_rows = (20, 60) -> settings__parallel_front_edge_rows_(20,60)
        """
        if self.parallel_front_edge_rows == None:
            return ""
        else:
            x0 = str(self.parallel_front_edge_rows[0])
            x1 = str(self.parallel_front_edge_rows[1])
            return f"__{conf.instance['notation']['settings_tags']['ci_mask']['parallel_front_edge_rows']}_({x0},{x1})"

    @property
    def parallel_trails_rows_tag(self):
        """Generate a parallel_trails_rows tag, to customize phase names based on the number of rows in the charge
        injection region in the trails of the parallel clocking direction are masked during the fit,

        This changes the phase settings folder as follows:

        parallel_trails_rows = None -> settings
        parallel_trails_rows = (0, 10) -> settings__parallel_trails_rows_(0,10)
        parallel_trails_rows = (20, 60) -> settings__parallel_trails_rows_(20,60)
        """
        if self.parallel_trails_rows == None:
            return ""
        else:
            x0 = str(self.parallel_trails_rows[0])
            x1 = str(self.parallel_trails_rows[1])
            return f"__{conf.instance['notation']['settings_tags']['ci_mask']['parallel_trails_rows']}_({x0},{x1})"

    @property
    def serial_front_edge_columns_tag(self):
        """Generate a serial_front_edge_columns tag, to customize phase names based on the number of columns in the
        charge  injection region at the front edge of the serial clocking direction are masked during the fit,

        This changes the phase settings folder as follows:

        serial_front_edge_columns = None -> settings
        serial_front_edge_columns = (0, 10) -> settings__serial_front_edge_columns_(0,10)
        serial_front_edge_columns = (20, 60) -> settings__serial_front_edge_columns_(20,60)
        """
        if self.serial_front_edge_columns == None:
            return ""
        else:
            x0 = str(self.serial_front_edge_columns[0])
            x1 = str(self.serial_front_edge_columns[1])
            return f"__{conf.instance['notation']['settings_tags']['ci_mask']['serial_front_edge_mask_rows']}_({x0},{x1})"

    @property
    def serial_trails_columns_tag(self):
        """Generate a serial_trails_columns tag, to customize phase names based on the number of columns in the charge
        injection region in the trails of the serial clocking direction are masked during the fit,

        This changes the phase settings folder as follows:

        serial_trails_columns = None -> settings
        serial_trails_columns = (0, 10) -> settings__serial_trails_columns_(0,10)
        serial_trails_columns = (20, 60) -> settings__serial_trails_columns_(20,60)
        """
        if self.serial_trails_columns == None:
            return ""
        else:
            x0 = str(self.serial_trails_columns[0])
            x1 = str(self.serial_trails_columns[1])
            return f"__{conf.instance['notation']['settings_tags']['ci_mask']['serial_trails_mask_rows']}_({x0},{x1})"


class CIMask(Mask2D):
    @classmethod
    def masked_front_edges_and_trails_from_ci_frame(cls, mask, ci_frame, settings):

        if settings.parallel_front_edge_rows is not None:

            parallel_front_edge_mask = cls.masked_parallel_front_edge_from_ci_frame(
                ci_frame=ci_frame, settings=settings
            )

            mask = mask + parallel_front_edge_mask

        if settings.parallel_trails_rows is not None:

            parallel_trails_mask = cls.masked_parallel_trails_from_ci_frame(
                ci_frame=ci_frame, settings=settings
            )

            mask = mask + parallel_trails_mask

        if settings.serial_front_edge_columns is not None:

            serial_front_edge_mask = cls.masked_serial_front_edge_from_ci_frame(
                ci_frame=ci_frame, settings=settings
            )

            mask = mask + serial_front_edge_mask

        if settings.serial_trails_columns is not None:

            serial_trails_mask = cls.masked_serial_trails_from_ci_frame(
                ci_frame=ci_frame, settings=settings
            )

            mask = mask + serial_trails_mask

        return mask

    @classmethod
    def masked_parallel_front_edge_from_ci_frame(cls, ci_frame, settings, invert=False):

        front_edge_regions = ci_frame.parallel_front_edge_regions(
            rows=settings.parallel_front_edge_rows
        )
        mask = np.full(ci_frame.shape_2d, False)

        for region in front_edge_regions:
            mask[region.y0 : region.y1, region.x0 : region.x1] = True

        if invert:
            mask = np.invert(mask)

        return CIMask(mask=mask.astype("bool"), pixel_scales=ci_frame.pixel_scales)

    @classmethod
    def masked_parallel_trails_from_ci_frame(cls, ci_frame, settings, invert=False):

        trails_regions = ci_frame.parallel_trails_regions(
            rows=settings.parallel_trails_rows
        )
        mask = np.full(ci_frame.shape_2d, False)

        for region in trails_regions:
            mask[region.y0 : region.y1, region.x0 : region.x1] = True

        if invert:
            mask = np.invert(mask)

        return CIMask(mask=mask.astype("bool"), pixel_scales=ci_frame.pixel_scales)

    @classmethod
    def masked_serial_front_edge_from_ci_frame(cls, ci_frame, settings, invert=False):

        front_edge_regions = ci_frame.serial_front_edge_regions(
            columns=settings.serial_front_edge_columns
        )
        mask = np.full(ci_frame.shape_2d, False)

        for region in front_edge_regions:
            mask[region.y0 : region.y1, region.x0 : region.x1] = True

        if invert:
            mask = np.invert(mask)

        return CIMask(mask=mask.astype("bool"), pixel_scales=ci_frame.pixel_scales)

    @classmethod
    def masked_serial_trails_from_ci_frame(cls, ci_frame, settings, invert=False):

        trails_regions = ci_frame.serial_trails_regions(
            columns=settings.serial_trails_columns
        )
        mask = np.full(ci_frame.shape_2d, False)

        for region in trails_regions:
            mask[region.y0 : region.y1, region.x0 : region.x1] = True

        if invert:
            mask = np.invert(mask)

        return CIMask(mask=mask.astype("bool"), pixel_scales=ci_frame.pixel_scales)



import copy

import numpy as np
from autoconf import conf
from autocti.charge_injection import ci_imaging, ci_fit, ci_hyper
from autofit.non_linear import abstract_search

from .hyper_phase import HyperPhase


class HyperNoisePhase(HyperPhase):
    def __init__(self, phase):

        super().__init__(phase=phase, hyper_name="hyper_noise")

    class Analysis(abstract_search.Analysis):
        def __init__(self, masked_ci_dataset_full, model_images):
            """
            An analysis to fit the noise for a single galaxy image.
            Parameters
            ----------
            masked_imaging: LensData
                Lens instrument, including an image and noise
            hyper_noise_scaling_map: np.ndarray
                An image produce of the overall system by a model
            hyper_noise_image_1d_path_dict: np.ndarray
                The contribution of one galaxy to the model image
            """

            self.masked_ci_dataset_full = masked_ci_dataset_full
            self.model_images = model_images

            # self.plot_hyper_noise_subplot = conf.instance.visualize.get(
            #     "plots", "plot_hyper_noise_subplot", bool
            # )

        def visualize(self, paths, instance, during_analysis):

            pass

        def log_likelihood_function(self, instance):
            """
            Fit the model image to the real image by scaling the hyper_noise noise.
            Parameters
            ----------
            instance: ModelInstance
                A model instance with a hyper_noise galaxy property
            Returns
            -------
            fit: float
            """

            hyper_noise_scalars = self.hyper_noise_scalars_from_instance(
                instance=instance
            )

            fits = list(
                map(
                    lambda data, model_image: self.fit_for_ci_data_model_image_and_hyper_noise_scalars(
                        ci_data=data,
                        model_image=model_image,
                        hyper_noise_scalars=hyper_noise_scalars,
                    ),
                    self.masked_ci_dataset_full,
                    self.model_images,
                )
            )

            return np.sum(list(map(lambda fit: fit.figure_of_merit, fits)))

        def hyper_noise_scalars_from_instance(self, instance):
            return [
                instance.hyper_noise_scalar_of_ci_regions,
                instance.hyper_noise_scalar_of_parallel_trails,
                instance.hyper_noise_scalar_of_serial_trails,
                instance.hyper_noise_scalar_of_serial_overscan_no_trails,
            ]

        def fit_for_ci_data_model_image_and_hyper_noise_scalars(
            self, ci_data, model_image, hyper_noise_scalars
        ):

            hyper_noise_map = ci_fit.hyper_noise_map_from_noise_map_and_noise_scalings(
                noise_scaling_maps=ci_data.noise_scaling_maps,
                hyper_noise_scalars=hyper_noise_scalars,
                noise_map=ci_data.noise_map,
            )

            return ci_fit.CIDataFit(
                masked_ci_imaging=ci_data,
                noise_map=hyper_noise_map,
                model_data=model_image,
            )

        @classmethod
        def describe(cls, instance):
            return "Running hyper_noise galaxy fit for HyperGalaxy:\n{}".format(
                instance.hyper_noise
            )

    def run_hyper(self, datasets, pool, results=None):
        """
        Run a fit for each galaxy from the previous phase.
        Parameters
        ----------
        simulator: LensData
        results: ResultsCollection
            Results from all previous phases
        Returns
        -------
        results: HyperGalaxyResults
            A collection of results, with one item per a galaxy
        """
        phase = self.make_hyper_phase()
        phase.hyper_noise_scalar_of_ci_regions = ci_hyper.CIHyperNoiseScalar
        phase.hyper_noise_scalar_of_parallel_trails = ci_hyper.CIHyperNoiseScalar
        phase.hyper_noise_scalar_of_serial_trails = ci_hyper.CIHyperNoiseScalar
        phase.hyper_noise_scalar_of_serial_overscan_no_trails = (
            ci_hyper.CIHyperNoiseScalar
        )

        masks = phase.mask_for_analysis_from_dataset(dataset=datasets)

        noise_scaling_maps_list = phase.noise_scaling_maps_list_from_total_images_and_results(
            total_images=len(datasets), results=results
        )

        masked_ci_dataset_full = list(
            map(
                lambda data, mask, maps: ci_imaging.MaskedCIImaging(
                    image=data.image,
                    noise_map=data.noise_map,
                    ci_pre_cti=data.ci_pre_cti,
                    mask=mask,
                    ci_pattern=data.ci_pattern,
                    ci_frame=data.ci_frame,
                    noise_scaling_maps_list=maps,
                ),
                datasets,
                masks,
                noise_scaling_maps_list,
            )
        )

        model_images = list(
            map(
                lambda max_log_likelihood_fit: max_log_likelihood_fit.model_image,
                results.last.max_log_likelihood_full_fits,
            )
        )

        hyper_result = copy.deepcopy(results.last)
        hyper_result.model = hyper_result.model.copy_with_fixed_priors(
            hyper_result.instance
        )

        hyper_result.analysis.uses_hyper_images = True
        hyper_result.analysis.model_images = model_images

        phase.search.model.parallel_trap = []
        phase.search.model.parallel_ccd = []
        phase.search.tag = ""

        phase.search.const_efficiency_mode = conf.instance.non_linear.get(
            "MultiNest", "const_efficiency_mode", bool
        )
        phase.search.facc = conf.instance.non_linear.get(
            "MultiNest", "sampling_efficiency", float
        )
        phase.search.n_live_points = conf.instance.non_linear.get(
            "MultiNest", "n_live_points", int
        )
        phase.search.multimodal = conf.instance.non_linear.get(
            "MultiNest", "multimodal", bool
        )

        analysis = self.Analysis(
            masked_ci_dataset_full=masked_ci_dataset_full, model_images=model_images
        )

        result = phase.search.f(analysis)

        def transfer_field(name):
            if hasattr(result.instance, name):
                setattr(hyper_result.instance, name, getattr(result.instance, name))
                setattr(hyper_result.model, name, getattr(result.model, name))

        transfer_field(name="hyper_noise_scalar_of_ci_regions")
        transfer_field(name="hyper_noise_scalar_of_parallel_trails")
        transfer_field(name="hyper_noise_scalar_of_serial_trails")
        transfer_field(name="hyper_noise_scalar_of_serial_overscan_no_trails")

        return hyper_result



"""
File: python/VIS_CTICalibrate/ChargeInjectPattern.py

Created on: 02/14/18
Author: James Nightingale
"""

from copy import deepcopy

import numpy as np
from autocti import exc
from autoarray.structures import region
from autoarray.util import frame_util
from autocti.charge_injection import ci_frame


class AbstractCIPattern(object):
    def __init__(self, normalization, regions):
        """ Abstract base class for a charge injection ci_pattern, which defines the regions charge injections appears \
         on a charge-injection ci_frame, the input normalization and other properties.

        Parameters
        -----------
        normalization : float
            The normalization of the charge injection lines.
        regions: [(int,)]
            A list of the integer coordinates specifying the corners of each charge injection region \
            (top-row, bottom-row, left-column, right-column).
        """
        self.normalization = normalization
        self.regions = list(map(region.Region, regions))

    def with_extracted_regions(self, extraction_region):

        ci_pattern = deepcopy(self)

        extracted_regions = list(
            map(
                lambda region: frame_util.region_after_extraction(
                    original_region=region, extraction_region=extraction_region
                ),
                self.regions,
            )
        )
        extracted_regions = list(filter(None, extracted_regions))
        if not extracted_regions:
            extracted_regions = None

        ci_pattern.regions = extracted_regions
        return ci_pattern

    def check_pattern_is_within_image_dimensions(self, dimensions):

        for region in self.regions:

            if region.y1 > dimensions[0] or region.x1 > dimensions[1]:
                raise exc.CIPatternException(
                    "The charge injection ci_pattern regions are bigger than the image image_shape"
                )

    @property
    def total_rows_min(self):
        return np.min(list(map(lambda region: region.total_rows, self.regions)))

    @property
    def total_columns_min(self):
        return np.min(list(map(lambda region: region.total_columns, self.regions)))

    @property
    def rows_between_regions(self):
        return [
            self.regions[i + 1].y0 - self.regions[i].y1
            for i in range(len(self.regions) - 1)
        ]


class CIPatternUniform(AbstractCIPattern):
    """ A uniform charge injection ci_pattern, which is defined by the regions it appears on the charge injection \
        ci_frame and its normalization.

    """

    def ci_pre_cti_from(self, shape_2d, pixel_scales):
        """Use this charge injection ci_pattern to generate a pre-cti charge injection image. This is performed by \
        going to its charge injection regions and adding the charge injection normalization value.

        Parameters
        -----------
        shape_2d : (int, int)
            The image_shape of the ci_pre_ctis to be created.
        """

        self.check_pattern_is_within_image_dimensions(shape_2d)

        ci_pre_cti = np.zeros(shape_2d)

        for region in self.regions:
            ci_pre_cti[region.slice] += self.normalization

        return ci_frame.CIFrame.manual(
            array=ci_pre_cti, ci_pattern=self, pixel_scales=pixel_scales
        )


class CIPatternNonUniform(AbstractCIPattern):
    def __init__(
        self,
        normalization,
        regions,
        row_slope,
        column_sigma=None,
        maximum_normalization=np.inf,
    ):
        """ A non-uniform charge injection ci_pattern, which is defined by the regions it appears on a charge injection
        ci_frame and its average normalization.

        Non-uniformity across the columns of a charge injection ci_pattern is due to spikes / drops in the current that
        injects the charge. This is a noisy process, leading to non-uniformity with no regularity / smoothness. Thus,
        it cannot be modeled with an analytic profile, and must be assumed as prior-knowledge about the charge
        injection electronics or estimated from the observed charge injection ci_data.

        Non-uniformity across the rows of a charge injection ci_pattern is due to a drop-off in voltage in the current.
        Therefore, it appears smooth and be modeled as an analytic function, which this code assumes is a
        power-law with slope row_slope.

        Parameters
        -----------
        normalization : float
            The normalization of the charge injection region.
        regions : [(int,)]
            A list of the integer coordinates specifying the corners of each charge injection region
            (top-row, bottom-row, left-column, right-column).
        row_slope : float
            The power-law slope of non-uniformity in the row charge injection profile.
        """
        super(CIPatternNonUniform, self).__init__(normalization, regions)
        self.row_slope = row_slope
        self.column_sigma = column_sigma
        self.maximum_normalization = maximum_normalization

    def ci_pre_cti_from(self, shape_2d, pixel_scales, ci_seed=-1):
        """Use this charge injection ci_pattern to generate a pre-cti charge injection image. This is performed by going \
        to its charge injection regions and adding its non-uniform charge distribution.

        For one column of a non-uniform charge injection ci_pre_ctis, it is assumed that each non-uniform charge \
        injection region has the same overall normalization value (after drawing this value randomly from a Gaussian \
        distribution). Physically, this is true provided the spikes / troughs in the current that cause \
        non-uniformity occur in an identical fashion for the generation of each charge injection region.

        Parameters
        -----------
        column_sigma
        shape_2d
            The image_shape of the ci_pre_ctis to be created.
        maximum_normalization

        ci_seed : int
            Input ci_seed for the random number generator to give reproducible results. A new ci_seed is always used for each \
            ci_pre_ctis, ensuring each non-uniform ci_region has the same column non-uniformity ci_pattern.
        """

        self.check_pattern_is_within_image_dimensions(shape_2d)

        ci_pre_cti = np.zeros(shape_2d)

        if ci_seed == -1:
            ci_seed = np.random.randint(
                0, int(1e9)
            )  # Use one ci_seed, so all regions have identical column
            # non-uniformity.

        for region in self.regions:
            ci_pre_cti[region.slice] += self.ci_region_from_region(
                region_dimensions=region.shape, ci_seed=ci_seed
            )

        return ci_frame.CIFrame.manual(
            array=ci_pre_cti, ci_pattern=self, pixel_scales=pixel_scales
        )

    def ci_region_from_region(self, region_dimensions, ci_seed):
        """Generate the non-uniform charge distribution of a charge injection region. This includes non-uniformity \
        across both the rows and columns of the charge injection region.

        Before adding non-uniformity to the rows and columns, we assume an input charge injection level \
        (e.g. the average current being injected). We then simulator non-uniformity in this region.

        Non-uniformity in the columns is caused by sharp peaks and troughs in the input charge current. To simulator  \
        this, we change the normalization of each column by drawing its normalization value from a Gaussian \
        distribution which has a mean of the input normalization and standard deviation *column_sigma*. The seed \
        of the random number generator ensures that the non-uniform charge injection update_via_regions of each ci_pre_ctis \
        are identical.

        Non-uniformity in the rows is caused by the charge smoothly decreasing as the injection is switched off. To \
        simulator this, we assume the charge level as a function of row number is not flat but defined by a \
        power-law with slope *row_slope*.

        Non-uniform charge injection images are generated using the function *simulate_pre_cti*, which uses this \
        function.

        Parameters
        -----------
        maximum_normalization
        column_sigma
        region_dimensions : (int, int)
            The size of the non-uniform charge injection region.
        ci_seed : int
            Input seed for the random number generator to give reproducible results.
        """

        np.random.seed(ci_seed)

        ci_rows = region_dimensions[0]
        ci_columns = region_dimensions[1]
        ci_region = np.zeros(region_dimensions)

        for column_number in range(ci_columns):

            column_normalization = 0
            while (
                column_normalization <= 0
                or column_normalization >= self.maximum_normalization
            ):
                column_normalization = np.random.normal(
                    self.normalization, self.column_sigma
                )

            ci_region[0:ci_rows, column_number] = self.generate_column(
                size=ci_rows, normalization=column_normalization
            )

        return ci_region

    def generate_column(self, size, normalization):
        """Generate a column of non-uniform charge, including row non-uniformity.

        The pixel-numbering used to generate non-uniformity across the charge injection rows runs from 1 -> size

        Parameters
        -----------
        size : int
            The size of the non-uniform column of charge
        normalization : float
            The input normalization of the column's charge e.g. the level of charge injected.

        """
        return normalization * (np.arange(1, size + 1)) ** self.row_slope



from autocti.pipeline.phase.settings import SettingsPhaseCIImaging
from autocti.charge_injection import ci_mask, ci_imaging
from autocti.pipeline.phase.dataset.phase import PhaseDataset
from autocti.pipeline.phase.ci_imaging.analysis import Analysis
from autocti.pipeline.phase.ci_imaging.result import Result
from autofit.non_linear.paths import convert_paths
from autofit.tools.phase_property import PhaseProperty


class PhaseCIImaging(PhaseDataset):

    hyper_noise_scalar_of_ci_regions = PhaseProperty("hyper_noise_scalar_of_ci_regions")
    hyper_noise_scalar_of_parallel_trails = PhaseProperty(
        "hyper_noise_scalar_of_parallel_trails"
    )
    hyper_noise_scalar_of_serial_trails = PhaseProperty(
        "hyper_noise_scalar_of_serial_trails"
    )
    hyper_noise_scalar_of_serial_overscan_no_trails = PhaseProperty(
        "hyper_noise_scalar_of_serial_overscan_no_trails"
    )

    Analysis = Analysis
    Result = Result

    def __init__(
        self,
        *,
        search,
        parallel_traps=None,
        parallel_ccd=None,
        serial_traps=None,
        serial_ccd=None,
        hyper_noise_scalar_of_ci_regions=None,
        hyper_noise_scalar_of_parallel_trails=None,
        hyper_noise_scalar_of_serial_trails=None,
        hyper_noise_scalar_of_serial_overscan_no_trails=None,
        settings=SettingsPhaseCIImaging(),
    ):

        """

        A phase in an cti pipeline. Uses the set non_linear search to try to fit models and hyper_galaxies
        passed to it.

        Parameters
        ----------
        search: class
            The class of a non_linear search
        sub_size: int
            The side length of the subgrid
        """

        super().__init__(
            parallel_traps=parallel_traps,
            parallel_ccd=parallel_ccd,
            serial_traps=serial_traps,
            serial_ccd=serial_ccd,
            settings=settings,
            search=search,
        )

        self.hyper_noise_scalar_of_ci_regions = hyper_noise_scalar_of_ci_regions
        self.hyper_noise_scalar_of_parallel_trails = (
            hyper_noise_scalar_of_parallel_trails
        )
        self.hyper_noise_scalar_of_serial_trails = hyper_noise_scalar_of_serial_trails
        self.hyper_noise_scalar_of_serial_overscan_no_trails = (
            hyper_noise_scalar_of_serial_overscan_no_trails
        )

    def make_analysis(self, datasets, clocker, results=None, pool=None):
        """
        Returns an analysis object. Also calls the prior passing and image modifying functions to allow child classes to
        change the behaviour of the phase.

        Parameters
        ----------
        clocker
        datasets
        pool
        results: [Results]
            The result from the previous phase

        Returns
        -------
        analysis: Analysis
            An analysis object that the `NonLinearSearch` calls to determine the fit of a set of values
        """

        mask = ci_mask.CIMask.unmasked(
            shape_2d=datasets[0].image.shape_2d, pixel_scales=datasets[0].image.shape_2d
        )

        masks = [
            self.mask_for_analysis_from_dataset(dataset=dataset, mask=mask)
            for dataset in datasets
        ]

        noise_scaling_maps_list = self.noise_scaling_maps_list_from_total_images_and_results(
            total_images=len(datasets), results=results
        )

        settings_masked_ci_imaging = self.settings.settings_masked_ci_imaging.modify_via_fit_type(
            is_parallel_fit=self.is_parallel_fit, is_serial_fit=self.is_serial_fit
        )

        masked_ci_imagings = [
            ci_imaging.MaskedCIImaging(
                ci_imaging=dataset,
                mask=mask,
                noise_scaling_maps=maps,
                settings=settings_masked_ci_imaging,
            )
            for dataset, mask, maps in zip(datasets, masks, noise_scaling_maps_list)
        ]

        return Analysis(
            masked_ci_imagings=masked_ci_imagings,
            clocker=clocker,
            settings_cti=self.settings.settings_cti,
            results=results,
            pool=pool,
        )

    def mask_for_analysis_from_dataset(self, dataset, mask):

        mask = self.mask_for_analysis_from_cosmic_ray_map(
            cosmic_ray_map=dataset.cosmic_ray_map, mask=mask
        )

        return ci_mask.CIMask.masked_front_edges_and_trails_from_ci_frame(
            mask=mask, ci_frame=dataset.image, settings=self.settings.settings_ci_mask
        )

    def noise_scaling_maps_list_from_total_images_and_results(
        self, total_images, results
    ):

        if self.model.hyper_noise_scalar_of_ci_regions is not None:
            noise_scaling_maps_list_of_ci_regions = (
                results.last.noise_scaling_maps_list_of_ci_regions
            )
        else:
            noise_scaling_maps_list_of_ci_regions = total_images * [None]

        if self.model.hyper_noise_scalar_of_parallel_trails is not None:
            noise_scaling_maps_list_of_parallel_trails = (
                results.last.noise_scaling_maps_list_of_parallel_trails
            )
        else:
            noise_scaling_maps_list_of_parallel_trails = total_images * [None]

        if self.model.hyper_noise_scalar_of_serial_trails is not None:
            noise_scaling_maps_list_of_serial_trails = (
                results.last.noise_scaling_maps_list_of_serial_trails
            )
        else:
            noise_scaling_maps_list_of_serial_trails = total_images * [None]

        if self.model.hyper_noise_scalar_of_serial_overscan_no_trails is not None:
            noise_scaling_maps_list_of_serial_overscan_no_trails = (
                results.last.noise_scaling_maps_list_of_serial_overscan_no_trails
            )
        else:
            noise_scaling_maps_list_of_serial_overscan_no_trails = total_images * [None]

        noise_scaling_maps_list = []

        for image_index in range(total_images):
            noise_scaling_maps_list.append(
                [
                    noise_scaling_maps_list_of_ci_regions[image_index],
                    noise_scaling_maps_list_of_parallel_trails[image_index],
                    noise_scaling_maps_list_of_serial_trails[image_index],
                    noise_scaling_maps_list_of_serial_overscan_no_trails[image_index],
                ]
            )

        for image_index in range(total_images):
            noise_scaling_maps_list[image_index] = [
                noise_scaling_map
                for noise_scaling_map in noise_scaling_maps_list[image_index]
                if noise_scaling_map is not None
            ]

        noise_scaling_maps_list = list(filter(None, noise_scaling_maps_list))

        if len(noise_scaling_maps_list) == 0:
            return total_images * [None]

        return noise_scaling_maps_list



from test_autocti.simulators import simulators

"""
Welcome to the PyAutoCTI dataset generator. Here, we'll make the datasets used to test and profile PyAutoCTI.

This consists of the following sets of images:

- An image where CTI is added with 1 parallel trap species.
- An image where CTI is added with 3 parallel trap species.
- An image where CTI is added with 1 serial trap species.
- An image where CTI is added with 3 serial trap species.
- An image where CTI is added with 1 parallel trap species and 1 serial trap species.
- An image where CTI is added with 3 parallel trap species and 3 serial trap species.

Each image is generated at 4 resolutions, 36x36 (Patch), 120x120 (Low_Res), 300x300 (Mid_Res), 600x600 (High_Res).
"""

resolutions = ["patch"]
normalizations = [1000.0, 10000.0, 25000.0, 84700.0]

# To simulate each galaxy, we pass it a resolution nd call its simulate function.

for resolution in resolutions:

    simulators.simulate__ci_uniform__parallel_x1(
        resolution=resolution, normalizations=normalizations
    )
    simulators.simulate__ci_uniform__parallel_x3(
        resolution=resolution, normalizations=normalizations
    )
    simulators.simulate__ci_uniform__serial_x1(
        resolution=resolution, normalizations=normalizations
    )
    simulators.simulate__ci_uniform__serial_x3(
        resolution=resolution, normalizations=normalizations
    )
    simulators.simulate__ci_uniform__parallel_x1__serial_x1(
        resolution=resolution, normalizations=normalizations
    )
    simulators.simulate__ci_uniform__parallel_x3__serial_x3(
        resolution=resolution, normalizations=normalizations
    )



import arcticpy


class CCD(arcticpy.CCD):
    def __init__(self, full_well_depth=1e4, well_notch_depth=0.0, well_fill_power=0.58):
        super().__init__(
            fraction_of_traps_per_phase=[1],
            full_well_depth=full_well_depth,
            well_fill_power=well_fill_power,
            well_notch_depth=well_notch_depth,
            well_bloom_level=None,
        )



from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)

from autoarray.util import array_util as array
from autoarray.util import frame_util as frame
from autoarray.util import fit_util as fit
from autoarray.util import mask_util as mask



from autofit.mock import mock
from autofit.tools.pipeline import ResultsCollection
from autofit.mock.mock import MockSearch, MockSamples
import numpy as np

### Mock AutoFit ###


class MockResult(mock.MockResult):
    def __init__(
        self,
        samples=None,
        instance=None,
        model=None,
        analysis=None,
        search=None,
        mask=None,
        model_image=None,
        noise_scaling_maps_list_of_ci_regions=None,
        noise_scaling_maps_list_of_parallel_trails=None,
        noise_scaling_maps_list_of_serial_trails=None,
        noise_scaling_maps_list_of_serial_overscan_no_trails=None,
        use_as_hyper_dataset=False,
    ):

        super().__init__(
            samples=samples,
            instance=instance,
            model=model,
            analysis=analysis,
            search=search,
        )

        self.previous_model = model
        self.gaussian_tuples = None
        self.mask = None
        self.positions = None
        self.mask = mask
        self.model_image = model_image
        self.noise_scaling_maps_list_of_ci_regions = (
            noise_scaling_maps_list_of_ci_regions
        )
        self.noise_scaling_maps_list_of_parallel_trails = (
            noise_scaling_maps_list_of_parallel_trails
        )
        self.noise_scaling_maps_list_of_serial_trails = (
            noise_scaling_maps_list_of_serial_trails
        )
        self.noise_scaling_maps_list_of_serial_overscan_no_trails = (
            noise_scaling_maps_list_of_serial_overscan_no_trails
        )
        self.hyper_combined = MockHyperCombinedPhase()
        self.use_as_hyper_dataset = use_as_hyper_dataset

    @property
    def last(self):
        return self


class MockResults(ResultsCollection):
    def __init__(
        self,
        samples=None,
        instance=None,
        model=None,
        analysis=None,
        search=None,
        mask=None,
        model_image=None,
        noise_scaling_maps_list_of_ci_regions=None,
        noise_scaling_maps_list_of_parallel_trails=None,
        noise_scaling_maps_list_of_serial_trails=None,
        noise_scaling_maps_list_of_serial_overscan_no_trails=None,
        use_as_hyper_dataset=False,
    ):
        """
        A collection of results from previous phases. Results can be obtained using an index or the name of the phase
        from whence they came.
        """

        super().__init__()

        result = MockResult(
            samples=samples,
            instance=instance,
            model=model,
            analysis=analysis,
            search=search,
            mask=mask,
            model_image=model_image,
            noise_scaling_maps_list_of_ci_regions=noise_scaling_maps_list_of_ci_regions,
            noise_scaling_maps_list_of_parallel_trails=noise_scaling_maps_list_of_parallel_trails,
            noise_scaling_maps_list_of_serial_trails=noise_scaling_maps_list_of_serial_trails,
            noise_scaling_maps_list_of_serial_overscan_no_trails=noise_scaling_maps_list_of_serial_overscan_no_trails,
            use_as_hyper_dataset=use_as_hyper_dataset,
        )

        self.__result_list = [result]

    @property
    def last(self):
        """
        The result of the last phase
        """
        if len(self.__result_list) > 0:
            return self.__result_list[-1]
        return None

    def __getitem__(self, item):
        """
        Get the result of a previous phase by index

        Parameters
        ----------
        item: int
            The index of the result

        Returns
        -------
        result: Result
            The result of a previous phase
        """
        return self.__result_list[item]

    def __len__(self):
        return len(self.__result_list)


class MockHyperCombinedPhase:
    def __init__(self):
        pass


class MockPattern(object):
    def __init__(self, regions=None):

        self.normalization = 10
        self.regions = regions
        self.total_rows = 2
        self.total_columns = 2


class MockGeometry(object):
    def __init__(self):
        super(MockGeometry, self).__init__()


class MockFrameGeometry(object):
    def __init__(self, value=1.0):
        self.value = value

    def add_cti(self, image, cti_params, clocker):
        return self.value * np.ones((2, 2))


class MockCIFrame(object):
    def __init__(self, value=1.0):

        self.ci_pattern = MockPattern()
        self.frame_geometry = MockFrameGeometry(value=value)
        self.value = value

    def ci_regions_from_array(self, array):
        return array[0:2, 0]

    def parallel_non_ci_regions_frame_from_frame(self, array):
        return array[0:2, 1]

    def serial_all_trails_frame_from_frame(self, array):
        return array[0, 0:2]

    def serial_overscan_no_trails_frame_from_frame(self, array):
        return array[1, 0:2]

    def parallel_front_edge_line_binned_over_columns_from_frame(
        self, array, rows=None, mask=None
    ):
        return np.array([1.0, 1.0, 2.0, 2.0])

    def parallel_trails_line_binned_over_columns_from_frame(
        self, array, rows=None, mask=None
    ):
        return np.array([1.0, 1.0, 2.0, 2.0])

    def serial_front_edge_line_binned_over_rows_from_frame(
        self, array, columns=None, mask=None
    ):
        return np.array([1.0, 1.0, 2.0, 2.0])

    def serial_trails_line_binned_over_rows_from_frame(
        self, array, columns=None, mask=None
    ):
        return np.array([1.0, 1.0, 2.0, 2.0])


class MockCIPreCTI(np.ndarray):
    def __new__(
        cls,
        array,
        frame_geometry=MockGeometry(),
        ci_pattern=MockPattern(),
        value=1.0,
        *args,
        **kwargs
    ):
        ci = np.array(array).view(cls)
        ci.frame_geometry = frame_geometry
        ci.ci_pattern = ci_pattern
        ci.value = value
        return ci

    def ci_post_cti_from_cti_params_and_settings(self, cti_params, clocker):
        return self.value * np.ones((2, 2))


class MockChInj(np.ndarray):
    def __new__(cls, array, geometry=None, ci_pattern=None, *args, **kwargs):
        ci = np.array(array).view(cls)
        ci.frame_geometry = geometry
        ci.ci_pattern = ci_pattern
        return ci



from autoarray.plot import plotters
from autocti.plot import ci_line_plots


@plotters.set_include_and_sub_plotter
@plotters.set_labels
def subplot_ci_fit(fit, include=None, sub_plotter=None):
    """Plot the model datas_ of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customized.

    Parameters
    -----------
    xyticksize
    ysize
    xsize
    titlesize
    cb_tick_labels
    cb_tick_values
    cb_pad
    cb_fraction
    cb_ticksize
    linscale
    linthresh
    norm_max
    norm_min
    norm
    cmap
    aspect
    figsize
    fit : autocti.lens.fitting.Fitter
        Class containing fit between the model datas_ and observed lens datas_ (including residual_map, chi_squared_map etc.)
    output_path : str
        The path where the datas_ is output if the output_type is a file format (e.g. png, fits)
    output_filename : str
        The name of the file that is output, if the output_type is a file format (e.g. png, fits)
    output_format : str
        How the datas_ is output. File formats (e.g. png, fits) output the datas_ to harddisk. 'show' displays the datas_ \
        in the python interpreter window.
    """

    number_subplots = 9

    sub_plotter.open_subplot_figure(number_subplots=number_subplots)

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=1)

    image(fit=fit, include=include, plotter=sub_plotter)

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=2)

    noise_map(fit=fit, include=include, plotter=sub_plotter)

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=3)

    signal_to_noise_map(fit=fit, include=include, plotter=sub_plotter)

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=4)

    ci_pre_cti(fit=fit, include=include, plotter=sub_plotter)

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=5)

    ci_post_cti(fit=fit, include=include, plotter=sub_plotter)

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=7)

    residual_map(fit=fit, include=include, plotter=sub_plotter)

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=8)

    chi_squared_map(fit=fit, include=include, plotter=sub_plotter)

    sub_plotter.output.subplot_to_figure()

    sub_plotter.figure.close()


@plotters.set_include_and_sub_plotter
@plotters.set_labels
def subplot_residual_maps(fits, include=None, sub_plotter=None):
    """Plot the model datas_ of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customied.

    """

    number_subplots = len(fits)

    sub_plotter.open_subplot_figure(number_subplots=number_subplots)

    for index, fit in enumerate(fits):

        sub_plotter.setup_subplot(
            number_subplots=number_subplots, subplot_index=index + 1
        )

        residual_map(fit=fit, include=include, plotter=sub_plotter)

    sub_plotter.output.subplot_to_figure()

    sub_plotter.figure.close()


@plotters.set_include_and_sub_plotter
@plotters.set_labels
def subplot_normalized_residual_maps(fits, include=None, sub_plotter=None):
    """Plot the model datas_ of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customied.

    """

    number_subplots = len(fits)

    sub_plotter.open_subplot_figure(number_subplots=number_subplots)

    for index, fit in enumerate(fits):

        sub_plotter.setup_subplot(
            number_subplots=number_subplots, subplot_index=index + 1
        )

        normalized_residual_map(fit=fit, include=include, plotter=sub_plotter)

    sub_plotter.output.subplot_to_figure()

    sub_plotter.figure.close()


@plotters.set_include_and_sub_plotter
@plotters.set_labels
def subplot_chi_squared_maps(fits, include=None, sub_plotter=None):
    """Plot the model datas_ of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customized.

    """

    number_subplots = len(fits)

    sub_plotter.open_subplot_figure(number_subplots=number_subplots)

    for index, fit in enumerate(fits):

        sub_plotter.setup_subplot(
            number_subplots=number_subplots, subplot_index=index + 1
        )

        chi_squared_map(fit=fit, include=include, plotter=sub_plotter)

    sub_plotter.output.subplot_to_figure()

    sub_plotter.figure.close()


def individuals(
    fit,
    plot_image=False,
    plot_noise_map=False,
    plot_signal_to_noise_map=False,
    plot_ci_pre_cti=False,
    plot_ci_post_cti=False,
    plot_residual_map=False,
    plot_normalized_residual_map=False,
    plot_chi_squared_map=False,
    include=None,
    plotter=None,
):

    if plot_image:
        image(fit=fit, include=include, plotter=plotter)

    if plot_noise_map:
        noise_map(fit=fit, include=include, plotter=plotter)

    if plot_signal_to_noise_map:
        signal_to_noise_map(fit=fit, include=include, plotter=plotter)

    if plot_ci_pre_cti:
        ci_pre_cti(fit=fit, include=include, plotter=plotter)

    if plot_ci_post_cti:
        ci_post_cti(fit=fit, include=include, plotter=plotter)

    if plot_residual_map:
        residual_map(fit=fit, include=include, plotter=plotter)

    if plot_normalized_residual_map:
        normalized_residual_map(fit=fit, include=include, plotter=plotter)

    if plot_chi_squared_map:
        chi_squared_map(fit=fit, include=include, plotter=plotter)


@plotters.set_include_and_sub_plotter
@plotters.set_labels
def subplot_fit_lines(fit, line_region, include=None, sub_plotter=None):
    """Plot the model datas_ of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customized.

    """

    number_subplots = 9

    sub_plotter.open_subplot_figure(number_subplots=number_subplots)

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=1)

    image_line(fit=fit, line_region=line_region, include=include, plotter=sub_plotter)

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=2)

    noise_map_line(
        fit=fit, line_region=line_region, include=include, plotter=sub_plotter
    )

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=3)

    signal_to_noise_map_line(
        fit=fit, line_region=line_region, include=include, plotter=sub_plotter
    )

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=4)

    ci_pre_cti_line(
        fit=fit, line_region=line_region, include=include, plotter=sub_plotter
    )

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=5)

    ci_post_cti_line(
        fit=fit, line_region=line_region, include=include, plotter=sub_plotter
    )

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=7)

    residual_map_line(
        fit=fit, line_region=line_region, include=include, plotter=sub_plotter
    )

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=8)

    chi_squared_map_line(
        fit=fit, line_region=line_region, include=include, plotter=sub_plotter
    )

    sub_plotter.output.subplot_to_figure()

    sub_plotter.figure.close()


@plotters.set_include_and_sub_plotter
@plotters.set_labels
def subplot_residual_map_lines(fits, line_region, include=None, sub_plotter=None):
    """Plot the model datas_ of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customized.
    """

    number_subplots = len(fits)

    sub_plotter.open_subplot_figure(number_subplots=number_subplots)

    for index, fit in enumerate(fits):
        sub_plotter.setup_subplot(
            number_subplots=number_subplots, subplot_index=index + 1
        )

        residual_map_line(
            fit=fit, line_region=line_region, include=include, plotter=sub_plotter
        )

    sub_plotter.output.subplot_to_figure()

    sub_plotter.figure.close()


@plotters.set_include_and_sub_plotter
@plotters.set_labels
def subplot_normalized_residual_map_lines(
    fits, line_region, include=None, sub_plotter=None
):
    """Plot the model datas_ of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customized.
    """

    number_subplots = len(fits)

    sub_plotter.open_subplot_figure(number_subplots=number_subplots)

    for index, fit in enumerate(fits):
        sub_plotter.setup_subplot(
            number_subplots=number_subplots, subplot_index=index + 1
        )

        normalized_residual_map_line(
            fit=fit, line_region=line_region, include=include, plotter=sub_plotter
        )

    sub_plotter.output.subplot_to_figure()

    sub_plotter.figure.close()


@plotters.set_include_and_sub_plotter
@plotters.set_labels
def subplot_chi_squared_map_lines(fits, line_region, include=None, sub_plotter=None):
    """Plot the model datas_ of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customized.
    """

    number_subplots = len(fits)

    sub_plotter.open_subplot_figure(number_subplots=number_subplots)

    for index, fit in enumerate(fits):

        sub_plotter.setup_subplot(
            number_subplots=number_subplots, subplot_index=index + 1
        )

        chi_squared_map_line(
            fit=fit, line_region=line_region, include=include, plotter=sub_plotter
        )

    sub_plotter.output.subplot_to_figure()

    sub_plotter.figure.close()


def individuals_lines(
    fit,
    line_region,
    plot_image=False,
    plot_noise_map=False,
    plot_signal_to_noise_map=False,
    plot_ci_pre_cti=False,
    plot_ci_post_cti=False,
    plot_residual_map=False,
    plot_normalized_residual_map=False,
    plot_chi_squared_map=False,
    include=None,
    plotter=None,
):

    if plot_image:
        image_line(fit=fit, line_region=line_region, include=include, plotter=plotter)

    if plot_noise_map:
        noise_map_line(
            fit=fit, line_region=line_region, include=include, plotter=plotter
        )

    if plot_signal_to_noise_map:
        signal_to_noise_map_line(
            fit=fit, line_region=line_region, include=include, plotter=plotter
        )

    if plot_ci_pre_cti:
        ci_pre_cti_line(
            fit=fit, line_region=line_region, include=include, plotter=plotter
        )

    if plot_ci_post_cti:
        ci_post_cti_line(
            fit=fit, line_region=line_region, include=include, plotter=plotter
        )

    if plot_residual_map:
        residual_map_line(
            fit=fit, line_region=line_region, include=include, plotter=plotter
        )

    if plot_normalized_residual_map:
        normalized_residual_map_line(
            fit=fit, line_region=line_region, include=include, plotter=plotter
        )

    if plot_chi_squared_map:
        chi_squared_map_line(
            fit=fit, line_region=line_region, include=include, plotter=plotter
        )


@plotters.set_include_and_plotter
@plotters.set_labels
def image(fit, include=None, plotter=None):
    """Plot the observed image of the ccd simulator.

    Set *autocti.simulator.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    image : CIFrame
        The image of the dataset.
    """

    plotter.plot_frame(
        frame=fit.image,
        include_origin=include.origin,
        include_parallel_overscan=include.parallel_overscan,
        include_serial_prescan=include.serial_prescan,
        include_serial_overscan=include.serial_overscan,
    )


@plotters.set_include_and_plotter
@plotters.set_labels
def noise_map(fit, include=None, plotter=None):
    """Plot the observed noise_map of the ccd simulator.

    Set *autocti.simulator.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    noise_map : CIFrame
        The noise_map of the dataset.
    """

    plotter.plot_frame(
        frame=fit.noise_map,
        include_origin=include.origin,
        include_parallel_overscan=include.parallel_overscan,
        include_serial_prescan=include.serial_prescan,
        include_serial_overscan=include.serial_overscan,
    )


@plotters.set_include_and_plotter
@plotters.set_labels
def signal_to_noise_map(fit, include=None, plotter=None):
    """Plot the observed signal_to_noise_map of the ccd simulator.

    Set *autocti.simulator.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    signal_to_noise_map : CIFrame
        The signal_to_noise_map of the dataset.
    """

    plotter.plot_frame(
        frame=fit.signal_to_noise_map,
        include_origin=include.origin,
        include_parallel_overscan=include.parallel_overscan,
        include_serial_prescan=include.serial_prescan,
        include_serial_overscan=include.serial_overscan,
    )


@plotters.set_include_and_plotter
@plotters.set_labels
def ci_pre_cti(fit, include=None, plotter=None):
    """Plot the observed ci_pre_cti of the ccd simulator.

    Set *autocti.simulator.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    ci_pre_cti : CIFrame
        The ci_pre_cti of the dataset.
    """

    plotter.plot_frame(
        frame=fit.ci_pre_cti,
        include_origin=include.origin,
        include_parallel_overscan=include.parallel_overscan,
        include_serial_prescan=include.serial_prescan,
        include_serial_overscan=include.serial_overscan,
    )


@plotters.set_include_and_plotter
@plotters.set_labels
def ci_post_cti(fit, include=None, plotter=None):
    """Plot the observed ci_post_cti of the ccd simulator.

    Set *autocti.simulator.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    ci_post_cti : CIFrame
        The ci_post_cti of the dataset.
    """

    plotter.plot_frame(
        frame=fit.ci_post_cti,
        include_origin=include.origin,
        include_parallel_overscan=include.parallel_overscan,
        include_serial_prescan=include.serial_prescan,
        include_serial_overscan=include.serial_overscan,
    )


@plotters.set_include_and_plotter
@plotters.set_labels
def residual_map(fit, include=None, plotter=None):
    """Plot the observed residual_map of the ccd simulator.

    Set *autocti.simulator.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    residual_map : CIFrame
        The residual_map of the dataset.
    """

    plotter.plot_frame(
        frame=fit.residual_map,
        include_origin=include.origin,
        include_parallel_overscan=include.parallel_overscan,
        include_serial_prescan=include.serial_prescan,
        include_serial_overscan=include.serial_overscan,
    )


@plotters.set_include_and_plotter
@plotters.set_labels
def normalized_residual_map(fit, include=None, plotter=None):
    """Plot the observed normalized_residual_map of the ccd simulator.

    Set *autocti.simulator.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    normalized_residual_map : CIFrame
        The normalized_residual_map of the dataset.
    """

    plotter.plot_frame(
        frame=fit.normalized_residual_map,
        include_origin=include.origin,
        include_parallel_overscan=include.parallel_overscan,
        include_serial_prescan=include.serial_prescan,
        include_serial_overscan=include.serial_overscan,
    )


@plotters.set_include_and_plotter
@plotters.set_labels
def chi_squared_map(fit, include=None, plotter=None):
    """Plot the observed chi_squared_map of the ccd simulator.

    Set *autocti.simulator.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    chi_squared_map : CIFrame
        The chi_squared_map of the dataset.
    """

    plotter.plot_frame(
        frame=fit.chi_squared_map,
        include_origin=include.origin,
        include_parallel_overscan=include.parallel_overscan,
        include_serial_prescan=include.serial_prescan,
        include_serial_overscan=include.serial_overscan,
    )


@plotters.set_include_and_sub_plotter
@plotters.set_labels
def noise_scaling_maps(fit, include=None, sub_plotter=None):
    """Plot the observed chi_squared_map of the ccd simulator.

    Set *autocti.simulator.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    chi_squared_map : CIFrame
        The chi_squared_map of the dataset.
    """

    number_subplots = len(fit.noise_scaling_maps)

    sub_plotter.open_subplot_figure(number_subplots=number_subplots)

    for index in range(len(fit.noise_scaling_maps)):

        sub_plotter.setup_subplot(
            number_subplots=number_subplots, subplot_index=index + 1
        )

        sub_plotter.plot_frame(
            frame=fit.noise_scaling_maps[index],
            include_origin=include.origin,
            include_parallel_overscan=include.parallel_overscan,
            include_serial_prescan=include.serial_prescan,
            include_serial_overscan=include.serial_overscan,
        )

    sub_plotter.output.subplot_to_figure()

    sub_plotter.figure.close()


@plotters.set_include_and_plotter
@plotters.set_labels
def image_line(fit, line_region, include=None, plotter=None):
    """Plot the observed image of the ccd simulator.

    Set *autocti.simulator.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    image : CIFrame
        The image of the dataset.
    """
    ci_line_plots.plot_line_from_ci_frame(
        ci_frame=fit.image, line_region=line_region, include=include, plotter=plotter
    )


@plotters.set_include_and_plotter
@plotters.set_labels
def noise_map_line(fit, line_region, include=None, plotter=None):
    """Plot the observed noise_map of the ccd simulator.

    Set *autocti.simulator.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    noise_map : CIFrame
        The noise_map of the dataset.
    """
    ci_line_plots.plot_line_from_ci_frame(
        ci_frame=fit.noise_map,
        line_region=line_region,
        include=include,
        plotter=plotter,
    )


@plotters.set_include_and_plotter
@plotters.set_labels
def signal_to_noise_map_line(fit, line_region, include=None, plotter=None):
    """Plot the observed signal_to_noise_map of the ccd simulator.

    Set *autocti.simulator.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    signal_to_noise_map : CIFrame
        The signal_to_noise_map of the dataset.
    """
    ci_line_plots.plot_line_from_ci_frame(
        ci_frame=fit.signal_to_noise_map,
        line_region=line_region,
        include=include,
        plotter=plotter,
    )


@plotters.set_include_and_plotter
@plotters.set_labels
def ci_pre_cti_line(fit, line_region, include=None, plotter=None):
    """Plot the observed ci_pre_cti of the ccd simulator.

    Set *autocti.simulator.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    ci_pre_cti : CIFrame
        The ci_pre_cti of the dataset.
    """
    ci_line_plots.plot_line_from_ci_frame(
        ci_frame=fit.ci_pre_cti,
        line_region=line_region,
        include=include,
        plotter=plotter,
    )


@plotters.set_include_and_plotter
@plotters.set_labels
def ci_post_cti_line(fit, line_region, include=None, plotter=None):
    """Plot the observed ci_post_cti of the ccd simulator.

    Set *autocti.simulator.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    ci_post_cti : CIFrame
        The ci_post_cti of the dataset.
    """
    ci_line_plots.plot_line_from_ci_frame(
        ci_frame=fit.ci_post_cti,
        line_region=line_region,
        include=include,
        plotter=plotter,
    )


@plotters.set_include_and_plotter
@plotters.set_labels
def residual_map_line(fit, line_region, include=None, plotter=None):
    """Plot the observed residual_map of the ccd simulator.

    Set *autocti.simulator.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    residual_map : CIFrame
        The residual_map of the dataset.
    """
    ci_line_plots.plot_line_from_ci_frame(
        ci_frame=fit.residual_map,
        line_region=line_region,
        include=include,
        plotter=plotter,
    )


@plotters.set_include_and_plotter
@plotters.set_labels
def normalized_residual_map_line(fit, line_region, include=None, plotter=None):
    """Plot the observed normalized_residual_map of the ccd simulator.

    Set *autocti.simulator.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    normalized_residual_map : CIFrame
        The normalized_residual_map of the dataset.
    """
    ci_line_plots.plot_line_from_ci_frame(
        ci_frame=fit.normalized_residual_map,
        line_region=line_region,
        include=include,
        plotter=plotter,
    )


@plotters.set_include_and_plotter
@plotters.set_labels
def chi_squared_map_line(fit, line_region, include=None, plotter=None):
    """Plot the observed chi_squared_map of the ccd simulator.

    Set *autocti.simulator.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    chi_squared_map : CIFrame
        The chi_squared_map of the dataset.
    """
    ci_line_plots.plot_line_from_ci_frame(
        ci_frame=fit.chi_squared_map,
        line_region=line_region,
        include=include,
        plotter=plotter,
    )



from autoconf import conf
from autoarray.mask import mask_2d
from autoarray.structures import abstract_structure
from autoarray import exc
from autoarray.structures import region as reg
from autoarray.util import array_util

import numpy as np


class SettingsMask:
    def __init__(
        self,
        cosmic_ray_parallel_buffer=10,
        cosmic_ray_serial_buffer=10,
        cosmic_ray_diagonal_buffer=3,
    ):

        self.cosmic_ray_parallel_buffer = cosmic_ray_parallel_buffer
        self.cosmic_ray_serial_buffer = cosmic_ray_serial_buffer
        self.cosmic_ray_diagonal_buffer = cosmic_ray_diagonal_buffer

    @property
    def tag(self):
        return (
            f"{conf.instance['notation']['settings_tags']['mask']['mask']}["
            f"{self.cosmic_ray_buffer_tag}]"
        )

    @property
    def cosmic_ray_buffer_tag(self):
        """Generate a cosmic ray buffer tag, to customize phase names based on the size of the cosmic ray masks in the \
        parallel, serial and diagonal directions

        This changes the phase settings folder as follows:

        cosmic_ray_parallel_buffer = 1, cosmic_ray_serial_buffer=2, cosmic_ray_diagonal_buffer=3 = -> settings__cr_p1s2d3
        cosmic_ray_parallel_buffer = 10, cosmic_ray_serial_buffer=5, cosmic_ray_diagonal_buffer=1 = -> settings__cr_p10s5d1
        """

        if (
            self.cosmic_ray_diagonal_buffer is None
            and self.cosmic_ray_serial_buffer is None
            and self.cosmic_ray_diagonal_buffer is None
        ):
            return ""

        if self.cosmic_ray_parallel_buffer is None:
            cosmic_ray_parallel_buffer_tag = ""
        else:
            cosmic_ray_parallel_buffer_tag = f"{conf.instance['notation']['settings_tags']['mask']['cosmic_ray_parallel_buffer']}{self.cosmic_ray_parallel_buffer}"

        if self.cosmic_ray_serial_buffer is None:
            cosmic_ray_serial_buffer_tag = ""
        else:
            cosmic_ray_serial_buffer_tag = f"{conf.instance['notation']['settings_tags']['mask']['cosmic_ray_serial_buffer']}{self.cosmic_ray_serial_buffer}"

        if self.cosmic_ray_diagonal_buffer is None:
            cosmic_ray_diagonal_buffer_tag = ""
        else:
            cosmic_ray_diagonal_buffer_tag = f"{conf.instance['notation']['settings_tags']['mask']['cosmic_ray_diagonal_buffer']}{self.cosmic_ray_diagonal_buffer}"

        return (
            f"__"
            f"{conf.instance['notation']['settings_tags']['mask']['cosmic_ray_buffer']}"
            f"_{cosmic_ray_parallel_buffer_tag}"
            f"{cosmic_ray_serial_buffer_tag}"
            f"{cosmic_ray_diagonal_buffer_tag}"
        )


class Mask2D(mask_2d.AbstractMask2D):
    @classmethod
    def manual(cls, mask, pixel_scales, origin=(0.0, 0.0), invert=False):
        """Create a Mask2D (see *Mask2D.__new__*) by inputting the array values in 2D, for example:

        mask=np.array([[False, False],
                       [True, False]])

        mask=[[False, False],
               [True, False]]

        Parameters
        ----------
        mask : np.ndarray or list
            The bool values of the mask input as an ndarray of shape [total_y_pixels, total_x_pixels ]or a list of
            lists.
        pixel_scales: (float, float) or float
            The (y,x) scaled units to pixel units conversion factors of every pixel. If this is input as a ``float``,
            it is converted to a (float, float) structure.
        origin : (float, float)
            The (y,x) scaled units origin of the mask's coordinate system.
        invert : bool
            If `True`, the ``bool``'s of the input ``mask`` are inverted, for example `False`'s become `True`
            and visa versa.
        """
        if type(mask) is list:
            mask = np.asarray(mask).astype("bool")

        if invert:
            mask = np.invert(mask)

        pixel_scales = abstract_structure.convert_pixel_scales(
            pixel_scales=pixel_scales
        )

        if len(mask.shape) != 2:
            raise exc.MaskException("The input mask is not a two dimensional array")

        return cls(mask=mask, pixel_scales=pixel_scales, origin=origin)

    @classmethod
    def unmasked(cls, shape_2d, pixel_scales, origin=(0.0, 0.0), invert=False):
        """Create a mask where all pixels are `False` and therefore unmasked.

        Parameters
        ----------
        mask : np.ndarray or list
            The bool values of the mask input as an ndarray of shape [total_y_pixels, total_x_pixels ]or a list of
            lists.
        pixel_scales: (float, float) or float
            The (y,x) scaled units to pixel units conversion factors of every pixel. If this is input as a ``float``,
            it is converted to a (float, float) structure.
        origin : (float, float)
            The (y,x) scaled units origin of the mask's coordinate system.
        invert : bool
            If `True`, the ``bool``'s of the input ``mask`` are inverted, for example `False`'s become `True`
            and visa versa.
        """
        return cls.manual(
            mask=np.full(shape=shape_2d, fill_value=False),
            pixel_scales=pixel_scales,
            origin=origin,
            invert=invert,
        )

    @classmethod
    def from_masked_regions(cls, shape_2d, pixel_scales, masked_regions):

        mask = cls.unmasked(shape_2d=shape_2d, pixel_scales=pixel_scales)
        masked_regions = list(
            map(lambda region: reg.Region(region=region), masked_regions)
        )
        for region in masked_regions:
            mask[region.y0 : region.y1, region.x0 : region.x1] = True

        return mask

    @classmethod
    def from_cosmic_ray_map_buffed(cls, cosmic_ray_map, settings=SettingsMask()):
        """
        Returns the mask used for CTI Calibration, which is all `False` unless specific regions are input for masking.

        Parameters
        ----------
        cosmic_ray_map : arrays.Array
            2D arrays flagging where cosmic rays on the image.
        cosmic_ray_parallel_buffer : int
            The number of pixels from each ray pixels are masked in the parallel direction.
        cosmic_ray_serial_buffer : int
            The number of pixels from each ray pixels are masked in the serial direction.
        cosmic_ray_diagonal_buffer : int
            The number of pixels from each ray pixels are masked in the digonal up from the parallel + serial direction.
        """
        mask = cls.unmasked(
            shape_2d=cosmic_ray_map.shape_2d, pixel_scales=cosmic_ray_map.pixel_scales
        )

        cosmic_ray_mask = (cosmic_ray_map > 0.0).astype("bool")

        for y in range(mask.shape[0]):
            for x in range(mask.shape[1]):
                if cosmic_ray_mask[y, x]:
                    y0, y1 = cosmic_ray_map.parallel_trail_from_y(
                        y=y, dy=settings.cosmic_ray_parallel_buffer
                    )
                    mask[y0:y1, x] = True
                    x0, x1 = cosmic_ray_map.serial_trail_from_x(
                        x=x, dx=settings.cosmic_ray_serial_buffer
                    )
                    mask[y, x0:x1] = True
                    y0, y1 = cosmic_ray_map.parallel_trail_from_y(
                        y=y, dy=settings.cosmic_ray_diagonal_buffer
                    )
                    x0, x1 = cosmic_ray_map.serial_trail_from_x(
                        x=x, dx=settings.cosmic_ray_diagonal_buffer
                    )
                    mask[y0:y1, x0:x1] = True

        return mask

    @classmethod
    def from_fits(
        cls, file_path, pixel_scales, hdu=0, origin=(0.0, 0.0), resized_mask_shape=None
    ):
        """
        Loads the image from a .fits file.

        Parameters
        ----------
        file_path : str
            The full path of the fits file.
        hdu : int
            The HDU number in the fits file containing the image image.
        pixel_scales : float or (float, float)
            The arc-second to pixel conversion factor of each pixel.
        """

        if type(pixel_scales) is not tuple:
            if type(pixel_scales) is float or int:
                pixel_scales = (float(pixel_scales), float(pixel_scales))

        mask = cls.manual(
            mask=array_util.numpy_array_2d_from_fits(file_path=file_path, hdu=hdu),
            pixel_scales=pixel_scales,
            origin=origin,
        )

        if resized_mask_shape is not None:
            mask = mask.resized_mask_from_new_shape(new_shape=resized_mask_shape)

        return mask



import os

from autoconf import conf
import autofit as af
from test_autocti.simulators import resolution_util


def run(
    module,
    test_name=None,
    search=af.DynestyStatic(),
    config_folder="config",
    clocker=None,
    pool=None,
):
    test_name = test_name or module.test_name
    test_path = "{}/../".format(os.path.dirname(os.path.realpath(__file__)))
    output_path = test_path + "output/"
    config_path = test_path + config_folder
    conf.instance = conf.Config(config_path=config_path, output_path=output_path)

    datasets = list(
        map(
            lambda normalization: resolution_util.load_test_ci_data(
                ci_data_type=module.ci_data_type,
                ci_data_model=module.ci_data_model,
                resolution=module.resolution,
                normalization=normalization,
            ),
            module.ci_normalizations,
        )
    )

    pipeline = module.make_pipeline(
        name=test_name, folders=[module.test_type, test_name], search=search
    )

    pipeline.run(datasets=datasets, clocker=clocker, pool=pool)


def run_a_mock(module):
    # noinspection PyTypeChecker
    run(
        module,
        test_name=f"{module.test_name}_mock",
        search=af.MockSearch,
        config_folder="config_mock",
    )



from autofit.non_linear import abstract_search


class Result(abstract_search.Result):
    def __init__(self, samples, previous_model, analysis, search):
        """
        The result of a phase
        """
        super().__init__(samples=samples, previous_model=previous_model, search=search)

        self.analysis = analysis
        self.search = search

    @property
    def clocker(self):
        return self.analysis.clocker



from autocti.pipeline.phase.abstract.result import Result
from autofit.non_linear.paths import convert_paths
from autofit.tools import phase as af_phase


# noinspection PyAbstractClass


class AbstractPhase(af_phase.AbstractPhase):

    Result = Result

    def __init__(self, *, search):
        """
        A phase in an lens pipeline. Uses the set non_linear search to try to fit
        models and hyper_galaxies passed to it.

        Parameters
        ----------
        search: class
            The class of a non_linear search
        name: str
            The name of this phase
        """

        super().__init__(search=search)

    def make_result(self, result, analysis):
        return self.Result(
            samples=result.samples,
            previous_model=result.model,
            analysis=analysis,
            search=self.search,
        )

    @property
    def is_parallel_fit(self):
        if self.model.parallel_ccd is not None and self.model.serial_ccd is None:
            return True
        else:
            return False

    @property
    def is_serial_fit(self):
        if self.model.parallel_ccd is None and self.model.serial_ccd is not None:
            return True
        else:
            return False

    @property
    def is_parallel_and_serial_fit(self):
        if self.model.parallel_ccd is not None and self.model.serial_ccd is not None:
            return True
        else:
            return False



def line_regions_from_plots(
    plot_parallel_front_edge_line,
    plot_parallel_trails_line,
    plot_serial_front_edge_line,
    plot_serial_trails_line,
):

    line_regions = []

    if plot_parallel_front_edge_line:
        line_regions.append("parallel_front_edge")

    if plot_parallel_trails_line:
        line_regions.append("parallel_trails")

    if plot_serial_front_edge_line:
        line_regions.append("serial_front_edge")

    if plot_serial_trails_line:
        line_regions.append("serial_trails")

    return line_regions



import copy
import logging

import numpy as np
from autoarray.dataset import imaging as im
from autocti.structures import arrays

logger = logging.getLogger(__name__)


class Imaging(im.AbstractImaging):
    def __init__(self, image, noise_map, name=None):
        """A collection of 2D imaging dataset(an image, noise-map, psf, etc.)

        Parameters
        ----------
        image : aa.Array
            The array of the image data, in units of electrons per second.
        noise_map : NoiseMap | float | ndarray
            An array describing the RMS standard deviation error in each pixel, preferably in units of electrons per
            second.
        background_noise_map : NoiseMap
            An array describing the RMS standard deviation error in each pixel due to the background sky noise_map,
            preferably in units of electrons per second.
        poisson_noise_map : NoiseMap
            An array describing the RMS standard deviation error in each pixel due to the Poisson counts of the source,
            preferably in units of electrons per second.
        exposure_time_map : aa.Array
            An array describing the effective exposure time in each imaging pixel.
        background_sky_map : aa.Scaled
            An array describing the background sky.
        """

        super(Imaging, self).__init__(image=image, noise_map=noise_map, name=name)

    @classmethod
    def from_fits(
        cls,
        image_path,
        pixel_scales,
        image_hdu=0,
        noise_map_path=None,
        noise_map_hdu=0,
        name=None,
    ):
        """Factory for loading the imaging data_type from .fits files, as well as computing properties like the noise-map,
        exposure-time map, etc. from the imaging-data.

        This factory also includes a number of routines for converting the imaging-data from unit_label not supported by PyAutoLens \
        (e.g. adus, electrons) to electrons per second.

        Parameters
        ----------
        name
        image_path : str
            The path to the image .fits file containing the image (e.g. '/path/to/image.fits')
        pixel_scales : float
            The size of each pixel in arc seconds.
        image_hdu : int
            The hdu the image is contained in the .fits file specified by *image_path*.
        noise_map_path : str
            The path to the noise_map .fits file containing the noise_map (e.g. '/path/to/noise_map.fits')
        noise_map_hdu : int
            The hdu the noise_map is contained in the .fits file specified by *noise_map_path*.
        """

        image = arrays.Array.from_fits(
            file_path=image_path, hdu=image_hdu, pixel_scales=pixel_scales
        )

        noise_map = arrays.Array.from_fits(
            file_path=noise_map_path, hdu=noise_map_hdu, pixel_scales=pixel_scales
        )

        return Imaging(image=image, noise_map=noise_map, name=name)

    def output_to_fits(self, image_path, noise_map_path=None, overwrite=False):

        self.image.output_to_fits(file_path=image_path, overwrite=overwrite)

        if self.noise_map is not None and noise_map_path is not None:
            self.noise_map.output_to_fits(file_path=noise_map_path, overwrite=overwrite)


class MaskedImaging(im.MaskedImaging):
    def __init__(self, imaging, mask):
        """
        The lens dataset is the collection of data_type (image, noise-map, PSF), a mask, grid, convolver \
        and other utilities that are used for modeling and fitting an image of a strong lens.

        Whilst the image, noise-map, etc. are loaded in 2D, the lens dataset creates reduced 1D arrays of each \
        for lens calculations.

        Parameters
        ----------
        imaging: im.Imaging
            The imaging data_type all in 2D (the image, noise-map, PSF, etc.)
        mask: msk.Mask2D
            The 2D mask that is applied to the image.
        """

        super().__init__(imaging=imaging, mask=mask)

        self.image = imaging.image * np.invert(mask)
        self.noise_map = imaging.noise_map * np.invert(mask)



import os
import shutil
from os import path

import pytest
import autofit as af
from autocti.pipeline import visualizer as vis
from autoconf import conf

directory = path.dirname(path.realpath(__file__))


@pytest.fixture(name="plot_path")
def make_visualizer_plotter_setup():
    return "{}/files/plot/visualizer".format(
        os.path.dirname(os.path.realpath(__file__))
    )


@pytest.fixture(autouse=True)
def set_config_path(plot_path):
    conf.instance.push(f"{directory}/config", output_path=plot_path)


class TestPhaseCIImagingVisualizer:
    def test__visualizes_imaging_using_configs(
        self, masked_ci_imaging_7x7, plot_path, plot_patch
    ):

        masked_ci_imaging_7x7.cosmic_ray_map[0, 0] = 1

        if os.path.exists(plot_path):
            shutil.rmtree(plot_path)

        visualizer = vis.PhaseCIImagingVisualizer(masked_dataset=masked_ci_imaging_7x7)

        visualizer.visualize_ci_imaging(paths=af.Paths())

        assert f"{plot_path}/image//subplots/subplot_ci_imaging.png" in plot_patch.paths
        assert f"{plot_path}/image//ci_imaging/image.png" in plot_patch.paths
        assert f"{plot_path}/image//ci_imaging/noise_map.png" not in plot_patch.paths
        assert (
            f"{plot_path}/image//ci_imaging/signal_to_noise_map.png"
            not in plot_patch.paths
        )
        assert f"{plot_path}/image//ci_imaging/ci_pre_cti.png" in plot_patch.paths
        assert f"{plot_path}/image//ci_imaging/cosmic_ray_map.png" in plot_patch.paths

    def test__visualizes_imaging_lines_using_configs(
        self, masked_ci_imaging_7x7, plot_path, plot_patch
    ):

        if os.path.exists(plot_path):
            shutil.rmtree(plot_path)

        visualizer = vis.PhaseCIImagingVisualizer(masked_dataset=masked_ci_imaging_7x7)

        visualizer.visualize_ci_imaging_lines(
            paths=af.Paths(), line_region="parallel_front_edge"
        )

        assert (
            f"{plot_path}/image//subplots/subplot_ci_lines_parallel_front_edge.png"
            in plot_patch.paths
        )
        assert (
            f"{plot_path}/image//ci_imaging_parallel_front_edge/image_line.png"
            in plot_patch.paths
        )
        assert (
            f"{plot_path}/image//ci_imaging_parallel_front_edge/noise_map_line.png"
            not in plot_patch.paths
        )
        assert (
            f"{plot_path}/image//ci_imaging_parallel_front_edge/signal_to_noise_map_line.png"
            not in plot_patch.paths
        )
        assert (
            f"{plot_path}/image//ci_imaging_parallel_front_edge/ci_pre_cti_line.png"
            in plot_patch.paths
        )

    def test___visualizes_fit_using_configs(
        self, masked_ci_imaging_7x7, ci_fit_7x7, plot_path, plot_patch
    ):

        if os.path.exists(plot_path):
            shutil.rmtree(plot_path)

        visualizer = vis.PhaseCIImagingVisualizer(masked_dataset=masked_ci_imaging_7x7)

        visualizer.visualize_ci_fit(
            paths=af.Paths(), fit=ci_fit_7x7, during_analysis=True
        )

        assert f"{plot_path}/image//subplots/subplot_ci_fit.png" in plot_patch.paths
        assert f"{plot_path}/image//fit_ci_imaging/image.png" in plot_patch.paths
        assert (
            f"{plot_path}/image//fit_ci_imaging/noise_map.png" not in plot_patch.paths
        )
        assert (
            f"{plot_path}/image//fit_ci_imaging/signal_to_noise_map.png"
            not in plot_patch.paths
        )
        assert f"{plot_path}/image//fit_ci_imaging/ci_pre_cti.png" in plot_patch.paths
        assert f"{plot_path}/image//fit_ci_imaging/ci_post_cti.png" in plot_patch.paths
        assert (
            f"{plot_path}/image//fit_ci_imaging/residual_map.png"
            not in plot_patch.paths
        )
        assert (
            f"{plot_path}/image//fit_ci_imaging/normalized_residual_map.png"
            in plot_patch.paths
        )
        assert (
            f"{plot_path}/image//fit_ci_imaging/chi_squared_map.png" in plot_patch.paths
        )

        visualizer.visualize_ci_fit(
            paths=af.Paths(), fit=ci_fit_7x7, during_analysis=False
        )

        assert f"{plot_path}/image//subplots/subplot_ci_fit.png" in plot_patch.paths
        assert f"{plot_path}/image//fit_ci_imaging/image.png" in plot_patch.paths
        assert f"{plot_path}/image//fit_ci_imaging/noise_map.png" in plot_patch.paths
        assert (
            f"{plot_path}/image//fit_ci_imaging/signal_to_noise_map.png"
            in plot_patch.paths
        )
        assert f"{plot_path}/image//fit_ci_imaging/ci_pre_cti.png" in plot_patch.paths
        assert f"{plot_path}/image//fit_ci_imaging/ci_post_cti.png" in plot_patch.paths
        assert f"{plot_path}/image//fit_ci_imaging/residual_map.png" in plot_patch.paths
        assert (
            f"{plot_path}/image//fit_ci_imaging/normalized_residual_map.png"
            in plot_patch.paths
        )
        assert (
            f"{plot_path}/image//fit_ci_imaging/chi_squared_map.png" in plot_patch.paths
        )

    def test___visualizes_fit_lines_using_configs(
        self, masked_ci_imaging_7x7, ci_fit_7x7, plot_path, plot_patch
    ):

        if os.path.exists(plot_path):
            shutil.rmtree(plot_path)

        visualizer = vis.PhaseCIImagingVisualizer(masked_dataset=masked_ci_imaging_7x7)

        visualizer.visualize_ci_fit_lines(
            paths=af.Paths(),
            fit=ci_fit_7x7,
            line_region="parallel_front_edge",
            during_analysis=True,
        )

        assert (
            f"{plot_path}/image//subplots/subplot_ci_fit_lines_parallel_front_edge.png"
            in plot_patch.paths
        )
        assert (
            f"{plot_path}/image//fit_ci_imaging_parallel_front_edge/image_line.png"
            in plot_patch.paths
        )
        assert (
            f"{plot_path}/image//fit_ci_imaging_parallel_front_edge/noise_map_line.png"
            not in plot_patch.paths
        )
        assert (
            plot_path
            + "fit_ci_imaging_parallel_front_edge/signal_to_noise_map_line.png"
            not in plot_patch.paths
        )
        assert (
            f"{plot_path}/image//fit_ci_imaging_parallel_front_edge/ci_pre_cti_line.png"
            in plot_patch.paths
        )
        assert (
            f"{plot_path}/image//fit_ci_imaging_parallel_front_edge/ci_post_cti_line.png"
            in plot_patch.paths
        )
        assert (
            f"{plot_path}/image//fit_ci_imaging_parallel_front_edge/residual_map_line.png"
            not in plot_patch.paths
        )
        assert (
            f"{plot_path}/image//fit_ci_imaging_parallel_front_edge/normalized_residual_map_line.png"
            in plot_patch.paths
        )
        assert (
            f"{plot_path}/image//fit_ci_imaging_parallel_front_edge/chi_squared_map_line.png"
            in plot_patch.paths
        )

        visualizer.visualize_ci_fit_lines(
            paths=af.Paths(),
            fit=ci_fit_7x7,
            line_region="parallel_front_edge",
            during_analysis=False,
        )

        assert (
            f"{plot_path}/image//subplots/subplot_ci_fit_lines_parallel_front_edge.png"
            in plot_patch.paths
        )
        assert (
            f"{plot_path}/image//fit_ci_imaging_parallel_front_edge/image_line.png"
            in plot_patch.paths
        )
        assert (
            f"{plot_path}/image//fit_ci_imaging_parallel_front_edge/noise_map_line.png"
            in plot_patch.paths
        )
        assert (
            f"{plot_path}/image//fit_ci_imaging_parallel_front_edge/signal_to_noise_map_line.png"
            in plot_patch.paths
        )
        assert (
            f"{plot_path}/image//fit_ci_imaging_parallel_front_edge/ci_pre_cti_line.png"
            in plot_patch.paths
        )
        assert (
            f"{plot_path}/image//fit_ci_imaging_parallel_front_edge/ci_post_cti_line.png"
            in plot_patch.paths
        )
        assert (
            f"{plot_path}/image//fit_ci_imaging_parallel_front_edge/residual_map_line.png"
            in plot_patch.paths
        )
        assert (
            f"{plot_path}/image//fit_ci_imaging_parallel_front_edge/normalized_residual_map_line.png"
            in plot_patch.paths
        )
        assert (
            f"{plot_path}/image//fit_ci_imaging_parallel_front_edge/chi_squared_map_line.png"
            in plot_patch.paths
        )

    def test___visualizes_multiple_ci_fits_subplot__using_configs(
        self, masked_ci_imaging_7x7, ci_fit_7x7, plot_path, plot_patch
    ):

        if os.path.exists(plot_path):
            shutil.rmtree(plot_path)

        visualizer = vis.PhaseCIImagingVisualizer(masked_dataset=masked_ci_imaging_7x7)

        visualizer.visualize_multiple_ci_fits_subplots(
            paths=af.Paths(), fits=[ci_fit_7x7]
        )

        assert (
            f"{plot_path}/image//subplots/subplot_residual_maps.png" in plot_patch.paths
        )
        assert (
            f"{plot_path}/image//subplots/subplot_normalized_residual_maps.png"
            in plot_patch.paths
        )
        assert (
            f"{plot_path}/image//subplots/subplot_chi_squared_maps.png"
            in plot_patch.paths
        )

    def test___visualizes_multiple_ci_fits_lines_subplot__using_configs(
        self, masked_ci_imaging_7x7, ci_fit_7x7, plot_path, plot_patch
    ):

        if os.path.exists(plot_path):
            shutil.rmtree(plot_path)

        visualizer = vis.PhaseCIImagingVisualizer(masked_dataset=masked_ci_imaging_7x7)

        visualizer.visualize_multiple_ci_fits_subplots_lines(
            paths=af.Paths(), fits=[ci_fit_7x7], line_region="parallel_front_edge"
        )

        assert (
            f"{plot_path}/image//subplots/subplot_residual_maps_lines_parallel_front_edge.png"
            in plot_patch.paths
        )
        assert (
            f"{plot_path}/image//subplots/subplot_normalized_residual_maps_lines_parallel_front_edge.png"
            in plot_patch.paths
        )
        assert (
            f"{plot_path}/image//subplots/subplot_chi_squared_maps_lines_parallel_front_edge.png"
            in plot_patch.paths
        )



import os

import autocti as ac
import autofit as af

test_path = "{}/../".format(os.path.dirname(os.path.realpath(__file__)))


def shape_2d_from_resolution(resolution):

    if resolution in "patch":
        return (36, 36)
    elif resolution in "lowres":
        return (100, 100)
    elif resolution in "midres":
        return (300, 300)
    elif resolution in "highres":
        return (600, 600)
    else:
        raise ValueError("An invalid simulator resolution was entered - ", resolution)


def ci_regions_from_resolution(resolution):

    if resolution in "patch":
        return [(1, 7, 1, 30), (17, 23, 1, 30)]
    elif resolution in "lowres":
        return [(10, 30, 10, 80), (60, 80, 10, 80)]
    elif resolution in "midres":
        return [(10, 40, 10, 280), (110, 140, 10, 280), (210, 240, 10, 280)]
    elif resolution in "highres":
        return [
            (10, 40, 10, 580),
            (110, 140, 10, 580),
            (210, 240, 10, 580),
            (310, 340, 10, 580),
            (410, 440, 10, 580),
            (510, 540, 10, 580),
        ]
    else:
        raise ValueError("An invalid simulator resolution was entered - ", resolution)


def resolution_from_shape(shape):

    if shape == (36, 36):
        return "patch"
    elif shape == (100, 100):
        return "lowres"
    elif shape == (300, 300):
        return "midres"
    elif shape == (600, 600):
        return "highres"
    else:
        raise ValueError(
            "An invalid shape was entered when generating the dataset-type - ", shape
        )


def simulate_ci_data_from_ci_normalization_region_and_cti_model(
    ci_data_type,
    ci_data_model,
    resolution,
    clocker,
    pattern,
    parallel_traps=None,
    parallel_ccd=None,
    serial_traps=None,
    serial_ccd=None,
    read_noise=1.0,
    cosmic_ray_map=None,
):

    shape = shape_2d_from_resolution(resolution=resolution)

    ci_pre_cti = pattern.ci_pre_cti_from(shape_2d=shape)

    simulator = ac.ci.SimulatorCIImaging(read_noise=read_noise)

    imaging = simulator.from_ci_pre_cti(
        clocker=clocker,
        ci_pre_cti=ci_pre_cti,
        ci_pattern=pattern,
        parallel_traps=parallel_traps,
        parallel_ccd=parallel_ccd,
        serial_traps=serial_traps,
        serial_ccd=serial_ccd,
        cosmic_ray_map=cosmic_ray_map,
    )

    # Now, lets output this simulated ccd-simulator to the test_autocti/simulator folder.
    test_path = "{}/../".format(os.path.dirname(os.path.realpath(__file__)))

    ci_data_path = af.util.create_path(
        path=test_path, folders=["dataset", ci_data_type, ci_data_model, resolution]
    )

    normalization = str(int(pattern.normalization))

    imaging.output_to_fits(
        image_path=ci_data_path + "image_" + normalization + ".fits",
        noise_map_path=ci_data_path + "noise_map_" + normalization + ".fits",
        ci_pre_cti_path=ci_data_path + "ci_pre_cti_" + normalization + ".fits",
        cosmic_ray_map_path=ci_data_path + "cosmic_ray_map_" + normalization + ".fits",
        overwrite=True,
    )


def load_test_ci_data(ci_data_type, ci_data_model, resolution, normalization):

    ci_regions = ci_regions_from_resolution(resolution=resolution)

    ci_pattern = ac.ci.CIPatternUniform(normalization=normalization, regions=ci_regions)

    dataset_path = af.util.create_path(
        path=test_path, folders=["dataset", ci_data_type, ci_data_model, resolution]
    )

    normalization = str(int(ci_pattern.normalization))

    if os.path.exists(f"{dataset_path}/cosmic_ray_map_{normalization}.fits"):
        cosmic_ray_map_path = f"{dataset_path}/cosmic_ray_map_{normalization}.fits"
    else:
        cosmic_ray_map_path = None

    return ac.ci.CIImaging.from_fits(
        roe_corner=(1, 0),
        ci_pattern=ci_pattern,
        pixel_scales=0.1,
        parallel_overscan=ac.Region((33, 36, 1, 30)),
        serial_overscan=ac.Region((0, 33, 30, 36)),
        serial_prescan=ac.Region((0, 36, 0, 1)),
        image_path=f"{dataset_path}/image_{normalization}.fits",
        noise_map_path=f"{dataset_path}/noise_map_{normalization}.fits",
        ci_pre_cti_path=f"{dataset_path}/ci_pre_cti_{normalization}.fits",
        cosmic_ray_map_path=cosmic_ray_map_path,
    )


# TODO : Keeping here for now so we have the overscans need to just load thm from data resolution.

# class CIFrame(ac.ci.CIFrame):
#     @classmethod
#     def patch(cls):
#         """This class represents the quadrant geometry of an integration quadrant."""
#         return cls.manual(
#             corner=(0, 0),
#             parallel_overscan=ac.Region((33, 36, 1, 30)),
#             serial_overscan=ac.Region((0, 33, 30, 36)),
#             serial_prescan=ac.Region((0, 36, 0, 1)),
#         )
#
#     @classmethod
#     def low_res(cls):
#         """This class represents the quadrant geometry of an integration quadrant."""
#         return cls.manual(
#             corner=(0, 0),
#             parallel_overscan=ac.Region((90, 100, 10, 80)),
#             serial_overscan=ac.Region((0, 90, 80, 100)),
#             serial_prescan=ac.Region((0, 100, 0, 10)),
#         )
#
#     @classmethod
#     def mid_res(cls):
#         """This class represents the quadrant geometry of an integration quadrant."""
#         return cls.manual(
#             corner=(0, 0),
#             parallel_overscan=ac.Region((280, 300, 10, 280)),
#             serial_overscan=ac.Region((0, 280, 280, 300)),
#             serial_prescan=ac.Region((0, 300, 0, 10)),
#         )
#
#     @classmethod
#     def high_res(cls):
#         """This class represents the quadrant geometry of an integration quadrant."""
#         return cls.manual(
#             corner=(0, 0),
#             parallel_overscan=ac.Region((580, 600, 20, 580)),
#             serial_overscan=ac.Region((0, 580, 580, 600)),
#             serial_prescan=ac.Region((0, 600, 20, 1)),
#         )



import os

import pytest
from autocti import plot as aplt

directory = os.path.dirname(os.path.realpath(__file__))


@pytest.fixture(name="plot_path")
def make_ci_fit_plotter_setup():
    return "{}/files/plots/fit/".format(os.path.dirname(os.path.realpath(__file__)))


def test__individual_attriute_plots__all_plot_correctly(
    ci_fit_7x7, plot_path, plot_patch
):

    aplt.CIFit.image(
        fit=ci_fit_7x7,
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )
    assert f"{plot_path}/image.png" in plot_patch.paths

    aplt.CIFit.noise_map(
        fit=ci_fit_7x7,
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )
    assert f"{plot_path}/noise_map.png" in plot_patch.paths

    aplt.CIFit.signal_to_noise_map(
        fit=ci_fit_7x7,
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )
    assert f"{plot_path}/signal_to_noise_map.png" in plot_patch.paths

    aplt.CIFit.ci_pre_cti(
        fit=ci_fit_7x7,
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )
    assert f"{plot_path}/ci_pre_cti.png" in plot_patch.paths

    aplt.CIFit.ci_post_cti(
        fit=ci_fit_7x7,
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )
    assert f"{plot_path}/ci_post_cti.png" in plot_patch.paths

    aplt.CIFit.residual_map(
        fit=ci_fit_7x7,
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )
    assert f"{plot_path}/residual_map.png" in plot_patch.paths

    aplt.CIFit.normalized_residual_map(
        fit=ci_fit_7x7,
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )
    assert f"{plot_path}/normalized_residual_map.png" in plot_patch.paths

    aplt.CIFit.chi_squared_map(
        fit=ci_fit_7x7,
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )
    assert f"{plot_path}/chi_squared_map.png" in plot_patch.paths

    aplt.CIFit.noise_scaling_maps(
        fit=ci_fit_7x7,
        sub_plotter=aplt.SubPlotter(output=aplt.Output(plot_path, format="png")),
    )

    assert f"{plot_path}/noise_scaling_maps.png" in plot_patch.paths


def test__individual_line_attriutes_plot__all_plot_correctly_output(
    ci_fit_7x7, plot_path, plot_patch
):

    aplt.CIFit.image_line(
        fit=ci_fit_7x7,
        line_region="parallel_front_edge",
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )

    assert f"{plot_path}/image_line.png" in plot_patch.paths

    aplt.CIFit.noise_map_line(
        fit=ci_fit_7x7,
        line_region="parallel_front_edge",
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )

    assert f"{plot_path}/noise_map_line.png" in plot_patch.paths

    aplt.CIFit.signal_to_noise_map_line(
        fit=ci_fit_7x7,
        line_region="parallel_front_edge",
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )

    assert f"{plot_path}/signal_to_noise_map_line.png" in plot_patch.paths

    aplt.CIFit.ci_pre_cti_line(
        fit=ci_fit_7x7,
        line_region="parallel_front_edge",
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )

    assert f"{plot_path}/ci_pre_cti_line.png" in plot_patch.paths

    aplt.CIFit.ci_post_cti_line(
        fit=ci_fit_7x7,
        line_region="parallel_front_edge",
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )

    assert f"{plot_path}/ci_post_cti_line.png" in plot_patch.paths

    aplt.CIFit.residual_map_line(
        fit=ci_fit_7x7,
        line_region="parallel_front_edge",
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )

    assert f"{plot_path}/residual_map_line.png" in plot_patch.paths

    aplt.CIFit.normalized_residual_map_line(
        fit=ci_fit_7x7,
        line_region="parallel_front_edge",
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )

    assert f"{plot_path}/normalized_residual_map_line.png" in plot_patch.paths

    aplt.CIFit.chi_squared_map_line(
        fit=ci_fit_7x7,
        line_region="parallel_front_edge",
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )

    assert f"{plot_path}/chi_squared_map_line.png" in plot_patch.paths


def test__ci_fit_subplots_are_output(ci_fit_7x7, plot_path, plot_patch):

    aplt.CIFit.subplot_ci_fit(
        fit=ci_fit_7x7,
        sub_plotter=aplt.SubPlotter(output=aplt.Output(plot_path, format="png")),
    )

    assert f"{plot_path}/subplot_ci_fit.png" in plot_patch.paths

    aplt.CIFit.subplot_residual_maps(
        fits=[ci_fit_7x7],
        sub_plotter=aplt.SubPlotter(output=aplt.Output(plot_path, format="png")),
    )

    assert f"{plot_path}/subplot_residual_maps.png" in plot_patch.paths

    aplt.CIFit.subplot_normalized_residual_maps(
        fits=[ci_fit_7x7],
        sub_plotter=aplt.SubPlotter(output=aplt.Output(plot_path, format="png")),
    )

    assert f"{plot_path}/subplot_normalized_residual_maps.png" in plot_patch.paths

    aplt.CIFit.subplot_chi_squared_maps(
        fits=[ci_fit_7x7],
        sub_plotter=aplt.SubPlotter(output=aplt.Output(plot_path, format="png")),
    )

    assert f"{plot_path}/subplot_chi_squared_maps.png" in plot_patch.paths


def test__ci_fit_subplots_lines_are_output(ci_fit_7x7, plot_path, plot_patch):

    aplt.CIFit.subplot_residual_map_lines(
        fits=[ci_fit_7x7],
        line_region="parallel_front_edge",
        sub_plotter=aplt.SubPlotter(output=aplt.Output(plot_path, format="png")),
    )

    assert f"{plot_path}/subplot_residual_map_lines.png" in plot_patch.paths

    aplt.CIFit.subplot_normalized_residual_map_lines(
        fits=[ci_fit_7x7],
        line_region="parallel_front_edge",
        sub_plotter=aplt.SubPlotter(output=aplt.Output(plot_path, format="png")),
    )

    assert f"{plot_path}/subplot_normalized_residual_map_lines.png" in plot_patch.paths

    aplt.CIFit.subplot_chi_squared_map_lines(
        fits=[ci_fit_7x7],
        line_region="parallel_front_edge",
        sub_plotter=aplt.SubPlotter(output=aplt.Output(plot_path, format="png")),
    )

    assert f"{plot_path}/subplot_chi_squared_map_lines.png" in plot_patch.paths


def test__fit_individuals__dependent_on_input(ci_fit_7x7, plot_path, plot_patch):

    aplt.CIFit.individuals(
        fit=ci_fit_7x7,
        plot_image=True,
        plot_noise_map=False,
        plot_signal_to_noise_map=False,
        plot_ci_pre_cti=True,
        plot_ci_post_cti=True,
        plot_chi_squared_map=True,
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )

    assert f"{plot_path}/image.png" in plot_patch.paths

    assert f"{plot_path}/noise_map.png" not in plot_patch.paths

    assert f"{plot_path}/signal_to_noise_map.png" not in plot_patch.paths

    assert f"{plot_path}/ci_pre_cti.png" in plot_patch.paths

    assert f"{plot_path}/ci_post_cti.png" in plot_patch.paths

    assert f"{plot_path}/residual_map.png" not in plot_patch.paths

    assert f"{plot_path}/chi_squared_map.png" in plot_patch.paths


def test__fit_individuals_line__dependent_on_input(ci_fit_7x7, plot_path, plot_patch):

    aplt.CIFit.individuals_lines(
        fit=ci_fit_7x7,
        line_region="parallel_front_edge",
        plot_image=True,
        plot_noise_map=False,
        plot_signal_to_noise_map=False,
        plot_ci_pre_cti=True,
        plot_ci_post_cti=True,
        plot_chi_squared_map=True,
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )

    assert f"{plot_path}/image_line.png" in plot_patch.paths

    assert f"{plot_path}/noise_map_line.png" not in plot_patch.paths

    assert f"{plot_path}/signal_to_noise_map_line.png" not in plot_patch.paths

    assert f"{plot_path}/ci_pre_cti_line.png" in plot_patch.paths

    assert f"{plot_path}/ci_post_cti_line.png" in plot_patch.paths

    assert f"{plot_path}/residual_map_line.png" not in plot_patch.paths

    assert f"{plot_path}/chi_squared_map_line.png" in plot_patch.paths



import autocti as ac
import autofit as af
from test_autocti.integration.tests import runner

test_type = "serial"
test_name = "x1_species__x2_image"
ci_data_type = "ci_uniform"
ci_data_model = "serial_x1"
resolution = "patch"
ci_normalizations = [10000.0, 84700.0]

clocker = ac.Clocker(serial_express=2)


def make_pipeline(name, folders, search=af.DynestyStatic()):

    serial_ccd = af.PriorModel(ac.CCD)

    serial_ccd.full_well_depth = 8.47e4
    serial_ccd.well_notch_depth = 1e-7

    phase1 = ac.PhaseCIImaging(
        name="phase_1",
        folders=folders,
        search=search,
        serial_traps=[af.PriorModel(ac.TrapInstantCapture)],
        serial_ccd=serial_ccd,
        settings=ac.SettingsPhaseCIImaging(rows=(0, 40)),
    )

    phase1 = phase1.extend_with_hyper_noise_phases()

    phase2 = ac.PhaseCIImaging(
        name="phase_2",
        folders=folders,
        serial_traps=phase1.result.model.serial_traps,
        serial_ccd=phase1.result.model.serial_ccd,
        hyper_noise_scalar_of_ci_regions=phase1.result.hyper_combined.instance.hyper_noise_scalar_of_ci_regions,
        hyper_noise_scalar_of_serial_trails=phase1.result.hyper_combined.instance.hyper_noise_scalar_of_serial_trails,
        search=search,
    )

    return ac.Pipeline(name, phase1, phase2)


if __name__ == "__main__":

    import sys

    runner.run(sys.modules[__name__], clocker=clocker)



import copy

import numpy as np
from autoconf import conf
from autocti.charge_injection import ci_frame, ci_pattern as pattern
from autoarray.dataset import preprocess, imaging
from autocti.mask import mask as msk
from autoarray.util import array_util
from autocti import exc


class CIImaging(imaging.AbstractImaging):
    def __init__(self, image, noise_map, ci_pre_cti, cosmic_ray_map=None, name=None):

        super().__init__(image=image, noise_map=noise_map, name=name)

        self.ci_pre_cti = ci_pre_cti
        self.cosmic_ray_map = cosmic_ray_map

    @property
    def mask(self):
        return msk.Mask2D.unmasked(
            shape_2d=self.shape_2d, pixel_scales=self.pixel_scales
        )

    @property
    def ci_pattern(self):
        return self.image.ci_pattern

    def parallel_calibration_ci_imaging_for_columns(self, columns):
        """
        Returnss a function to extract a parallel section for given columns
        """

        cosmic_ray_map = (
            self.cosmic_ray_map.parallel_calibration_frame_from_columns(columns=columns)
            if self.cosmic_ray_map is not None
            else None
        )

        return CIImaging(
            image=self.image.parallel_calibration_frame_from_columns(columns=columns),
            noise_map=self.noise_map.parallel_calibration_frame_from_columns(
                columns=columns
            ),
            ci_pre_cti=self.ci_pre_cti.parallel_calibration_frame_from_columns(
                columns=columns
            ),
            cosmic_ray_map=cosmic_ray_map,
        )

    def serial_calibration_ci_imaging_for_rows(self, rows):
        """
        Returnss a function to extract a serial section for given rows
        """

        cosmic_ray_map = (
            self.cosmic_ray_map.serial_calibration_frame_from_rows(rows=rows)
            if self.cosmic_ray_map is not None
            else None
        )

        return CIImaging(
            image=self.image.serial_calibration_frame_from_rows(rows=rows),
            noise_map=self.noise_map.serial_calibration_frame_from_rows(rows=rows),
            ci_pre_cti=self.ci_pre_cti.serial_calibration_frame_from_rows(rows=rows),
            cosmic_ray_map=cosmic_ray_map,
        )

    @classmethod
    def from_fits(
        cls,
        roe_corner,
        ci_pattern,
        image_path,
        pixel_scales,
        scans=None,
        image_hdu=0,
        noise_map_path=None,
        noise_map_hdu=0,
        noise_map_from_single_value=None,
        ci_pre_cti_path=None,
        ci_pre_cti_hdu=0,
        cosmic_ray_map_path=None,
        cosmic_ray_map_hdu=0,
    ):

        ci_image = ci_frame.CIFrame.from_fits(
            file_path=image_path,
            hdu=image_hdu,
            roe_corner=roe_corner,
            ci_pattern=ci_pattern,
            pixel_scales=pixel_scales,
            scans=scans,
        )

        if noise_map_path is not None:
            ci_noise_map = array_util.numpy_array_2d_from_fits(
                file_path=noise_map_path, hdu=noise_map_hdu
            )
        else:
            ci_noise_map = np.ones(ci_image.shape_2d) * noise_map_from_single_value

        ci_noise_map = ci_frame.CIFrame.manual(
            array=ci_noise_map,
            roe_corner=roe_corner,
            ci_pattern=ci_pattern,
            pixel_scales=pixel_scales,
            scans=scans,
        )

        if ci_pre_cti_path is not None:
            ci_pre_cti = array_util.numpy_array_2d_from_fits(
                file_path=ci_pre_cti_path, hdu=ci_pre_cti_hdu
            )
        else:
            if isinstance(ci_pattern, pattern.CIPatternUniform):
                ci_pre_cti = ci_pattern.ci_pre_cti_from(
                    shape_2d=ci_image.shape, pixel_scales=pixel_scales
                )
            else:
                raise exc.CIPatternException(
                    "Cannot estimate ci_pre_cti image from non-uniform charge injectiono pattern"
                )

        ci_pre_cti = ci_frame.CIFrame.manual(
            array=ci_pre_cti,
            roe_corner=roe_corner,
            ci_pattern=ci_pattern,
            pixel_scales=pixel_scales,
            scans=scans,
        )

        if cosmic_ray_map_path is not None:

            cosmic_ray_map = ci_frame.CIFrame.from_fits(
                file_path=cosmic_ray_map_path,
                hdu=cosmic_ray_map_hdu,
                roe_corner=roe_corner,
                ci_pattern=ci_pattern,
                pixel_scales=pixel_scales,
                scans=scans,
            )

        else:
            cosmic_ray_map = None

        return CIImaging(
            image=ci_image,
            noise_map=ci_noise_map,
            ci_pre_cti=ci_pre_cti,
            cosmic_ray_map=cosmic_ray_map,
        )

    def output_to_fits(
        self,
        image_path,
        noise_map_path=None,
        ci_pre_cti_path=None,
        cosmic_ray_map_path=None,
        overwrite=False,
    ):

        array_util.numpy_array_2d_to_fits(
            array_2d=self.image, file_path=image_path, overwrite=overwrite
        )

        if self.noise_map is not None and noise_map_path is not None:
            array_util.numpy_array_2d_to_fits(
                array_2d=self.noise_map, file_path=noise_map_path, overwrite=overwrite
            )

        if self.ci_pre_cti is not None and ci_pre_cti_path is not None:
            array_util.numpy_array_2d_to_fits(
                array_2d=self.ci_pre_cti, file_path=ci_pre_cti_path, overwrite=overwrite
            )

        if self.cosmic_ray_map is not None and cosmic_ray_map_path is not None:
            array_util.numpy_array_2d_to_fits(
                array_2d=self.cosmic_ray_map,
                file_path=cosmic_ray_map_path,
                overwrite=overwrite,
            )


class SettingsMaskedCIImaging(imaging.AbstractSettingsMaskedImaging):
    def __init__(self, parallel_columns=None, serial_rows=None):

        super().__init__()

        self.parallel_columns = parallel_columns
        self.serial_rows = serial_rows

    @property
    def tag(self):
        return (
            f"{conf.instance['notation']['settings_tags']['ci_imaging']['ci_imaging']}["
            f"{self.parallel_columns_tag}"
            f"{self.serial_rows_tag}]"
        )

    @property
    def parallel_columns_tag(self):
        """Generate a columns tag, to customize phase names based on the number of columns of simulator extracted in the fit,
        which is used to speed up parallel CTI fits.

        This changes the phase settings folder as follows:

        columns = None -> settings
        columns = 10 -> settings__col_10
        columns = 60 -> settings__col_60
        """
        if self.parallel_columns == None:
            return ""
        else:
            x0 = str(self.parallel_columns[0])
            x1 = str(self.parallel_columns[1])
            return f"__{conf.instance['notation']['settings_tags']['ci_imaging']['parallel_columns']}_({x0},{x1})"

    @property
    def serial_rows_tag(self):
        """Generate a rows tag, to customize phase names based on the number of rows of simulator extracted in the fit,
        which is used to speed up serial CTI fits.

        This changes the phase settings folder as follows:

        rows = None -> settings
        rows = (0, 10) -> settings__rows_(0,10)
        rows = (20, 60) -> settings__rows_(20,60)
        """
        if self.serial_rows == None:
            return ""
        else:
            x0 = str(self.serial_rows[0])
            x1 = str(self.serial_rows[1])
            return f"__{conf.instance['notation']['settings_tags']['ci_imaging']['serial_rows']}_({x0},{x1})"

    def modify_via_fit_type(self, is_parallel_fit, is_serial_fit):
        """Modify the settings based on the type of fit being performed where:

        - If the fit is a parallel only fit (is_parallel_fit=True, is_serial_fit=False) the serial_rows are set to None
          and all other settings remain the same.

        - If the fit is a serial only fit (is_parallel_fit=False, is_serial_fit=True) the parallel_columns are set to
          None and all other settings remain the same.

        - If the fit is a parallel and serial fit (is_parallel_fit=True, is_serial_fit=True) the *parallel_columns* and
          *serial_rows* are set to None and all other settings remain the same.

         These settings reflect the appropriate way to extract the charge injection imaging data for fits which use a
         parallel only CTI model, serial only CTI model or fit both.

         Parameters
         ----------
         is_parallel_fit : bool
            If True, the CTI model that is used to fit the charge injection data includes a parallel CTI component.
         is_serial_fit : bool
            If True, the CTI model that is used to fit the charge injection data includes a serial CTI component.
          """

        settings = copy.copy(self)

        if is_parallel_fit:
            settings.serial_rows = None

        if is_serial_fit:
            settings.parallel_columns = None

        return settings


class MaskedCIImaging(imaging.AbstractMaskedImaging):
    def __init__(
        self,
        ci_imaging,
        mask,
        noise_scaling_maps=None,
        settings=SettingsMaskedCIImaging(),
    ):
        """A data is the collection of simulator components (e.g. the image, noise-maps, PSF, etc.) which are used \
        to generate and fit it with a model image.

        The data is in 2D and masked, primarily to remove cosmic rays.

        The data also includes a number of attributes which are used to performt the fit, including (y,x) \
        grids of coordinates, convolvers and other utilities.

        Parameters
        ----------
        image : im.Image
            The 2D observed image and other observed quantities (noise-map, PSF, exposure-time map, etc.)
        mask: msk.Mask2D | None
            The 2D mask that is applied to image simulator.

        Attributes
        ----------
        image : ScaledSquarePixelArray
            The 2D observed image simulator (not an instance of im.Image, so does not include the other simulator attributes,
            which are explicitly made as new attributes of the data).
        noise_map : NoiseMap
            An arrays describing the RMS standard deviation error in each pixel, preferably in units of electrons per
            second.
        mask: msk.Mask2D
            The 2D mask that is applied to image simulator.
        """

        self.ci_imaging_full = copy.deepcopy(ci_imaging)
        self.ci_imaging_full.noise_scaling_maps = noise_scaling_maps
        self.mask_full = copy.deepcopy(mask)

        if settings.parallel_columns is not None:

            ci_imaging = self.ci_imaging_full.parallel_calibration_ci_imaging_for_columns(
                columns=settings.parallel_columns
            )

            mask = self.ci_imaging_full.image.parallel_calibration_mask_from_mask_and_columns(
                mask=mask, columns=settings.parallel_columns
            )

            if noise_scaling_maps is not None:
                noise_scaling_maps = [
                    noise_scaling_map.parallel_calibration_frame_from_columns(
                        columns=settings.parallel_columns
                    )
                    for noise_scaling_map in noise_scaling_maps
                ]

        if settings.serial_rows is not None:

            ci_imaging = self.ci_imaging_full.serial_calibration_ci_imaging_for_rows(
                rows=settings.serial_rows
            )

            mask = self.ci_imaging_full.image.serial_calibration_mask_from_mask_and_rows(
                mask=mask, rows=settings.serial_rows
            )

            if noise_scaling_maps is not None:

                noise_scaling_maps = [
                    noise_scaling_map.serial_calibration_frame_from_rows(
                        rows=settings.serial_rows
                    )
                    for noise_scaling_map in noise_scaling_maps
                ]

        super().__init__(imaging=ci_imaging, mask=mask, settings=settings)

        self.image = ci_frame.CIFrame.from_ci_frame(
            ci_frame=ci_imaging.image, mask=mask
        )
        self.noise_map = ci_frame.CIFrame.from_ci_frame(
            ci_frame=ci_imaging.noise_map, mask=mask
        )
        self.ci_pre_cti = ci_imaging.ci_pre_cti

        if ci_imaging.cosmic_ray_map is not None:

            self.cosmic_ray_map = ci_frame.CIFrame.from_ci_frame(
                ci_frame=ci_imaging.cosmic_ray_map, mask=mask
            )

        else:

            self.cosmic_ray_map = None

        if noise_scaling_maps is not None:
            self.noise_scaling_maps = [
                ci_frame.CIFrame.from_ci_frame(ci_frame=noise_scaling_map, mask=mask)
                for noise_scaling_map in noise_scaling_maps
            ]
        else:
            self.noise_scaling_maps = None

    @property
    def ci_imaging(self):
        return self.imaging


class SimulatorCIImaging(imaging.AbstractSimulatorImaging):
    def __init__(
        self,
        shape_2d,
        pixel_scales,
        read_noise=None,
        add_poisson_noise=False,
        scans=None,
        noise_if_add_noise_false=0.1,
        noise_seed=-1,
        ci_seed=-1,
    ):
        """A class representing a Imaging observation, using the shape of the image, the pixel scale,
        psf, exposure time, etc.

        Parameters
        ----------
        exposure_time_map : float
            The exposure time of an observation using this data_type.
        """

        super(SimulatorCIImaging, self).__init__(
            read_noise=read_noise,
            exposure_time=1.0,
            add_poisson_noise=add_poisson_noise,
            noise_if_add_noise_false=noise_if_add_noise_false,
            noise_seed=noise_seed,
        )

        self.shape_2d = shape_2d
        self.scans = scans
        self.pixel_scales = pixel_scales
        self.ci_seed = ci_seed

    def from_ci_pattern(
        self,
        ci_pattern,
        clocker,
        parallel_traps=None,
        parallel_ccd=None,
        serial_traps=None,
        serial_ccd=None,
        cosmic_ray_map=None,
        name=None,
    ):
        """Simulate a charge injection image, including effects like noises.

        Parameters
        -----------
        ci_pre_cti
        cosmic_ray_map
            The dimensions of the output simulated charge injection image.
        frame_geometry : ci_frame.CIQuadGeometry
            The quadrant geometry of the simulated image, defining where the parallel / serial overscans are and \
            therefore the direction of clocking and rotations before input into the cti algorithm.
        ci_pattern : ci_pattern.CIPatternSimulate
            The charge injection ci_pattern (regions, normalization, etc.) of the charge injection image.
        cti_params : ArcticParams.ArcticParams
            The CTI model parameters (trap density, trap release_timescales etc.).
        clocker : ArcticSettings.ArcticSettings
            The settings that control the cti clocking algorithm (e.g. ccd well_depth express option).
        read_noise : None or float
            The FWHM of the Gaussian read-noises added to the image.
        noise_seed : int
            Seed for the read-noises added to the image.
        """
        if isinstance(ci_pattern, pattern.CIPatternUniform):
            ci_pre_cti = ci_pattern.ci_pre_cti_from(
                shape_2d=self.shape_2d, pixel_scales=self.pixel_scales
            )
        else:
            ci_pre_cti = ci_pattern.ci_pre_cti_from(
                shape_2d=self.shape_2d,
                ci_seed=self.ci_seed,
                pixel_scales=self.pixel_scales,
            )

        if cosmic_ray_map is not None:
            ci_pre_cti += cosmic_ray_map

        ci_post_cti = clocker.add_cti(
            image=ci_pre_cti,
            parallel_traps=parallel_traps,
            parallel_ccd=parallel_ccd,
            serial_traps=serial_traps,
            serial_ccd=serial_ccd,
        )

        if self.read_noise is not None:
            ci_image = preprocess.data_with_gaussian_noise_added(
                data=ci_post_cti, sigma=self.read_noise, seed=self.noise_seed
            )
            ci_noise_map = self.read_noise * np.ones(ci_post_cti.shape)
        else:
            ci_image = ci_post_cti
            ci_noise_map = None

        return CIImaging(
            image=ci_frame.CIFrame.manual(
                array=ci_image,
                ci_pattern=ci_pattern,
                scans=self.scans,
                pixel_scales=self.pixel_scales,
            ),
            noise_map=ci_frame.CIFrame.manual(
                array=ci_noise_map,
                ci_pattern=ci_pattern,
                scans=self.scans,
                pixel_scales=self.pixel_scales,
            ),
            ci_pre_cti=ci_frame.CIFrame.manual(
                array=ci_pre_cti,
                ci_pattern=ci_pattern,
                scans=self.scans,
                pixel_scales=self.pixel_scales,
            ),
            cosmic_ray_map=ci_frame.CIFrame.manual(
                array=cosmic_ray_map,
                ci_pattern=ci_pattern,
                scans=self.scans,
                pixel_scales=self.pixel_scales,
            ),
            name=name,
        )



from autoarray.instruments import euclid
from autoarray.util import frame_util

from autocti.util import clocker as clk
from autocti.util import ccd
from autocti.util import traps

from autocti.charge_injection import ci_frame, ci_pattern as pattern
from autocti.charge_injection import ci_imaging

"""
Note on the rotations of frames:

The function 'non_uniform_frame_for_ou_sim' returns a frame that is rotated according to the ccd_id and quadrant id,
which given the quadrant unique define the orientation of the ndarray necessary to add paralle and serial
CTI in the correct direction. These are defined accoridng to:

    http://euclid.esac.esa.int/dm/dpdd/latest/le1dpd/dpcards/le1_visrawframe.html

This function will return a frame, which OU-Sim will then add the following effects too:

    - Add cosmic rays.
    - Bias.
    - Non-linearity.
    - Crosstalk?

The addition of CTI can then be performed using either the 'add_cti_to_frame_for_ou_sim' function, or the standard
function OU-Sim use to add CTI (I guess they ultimately both flow through arctic in an identical way, albeit our
code uses SWIG and omits the need to define arCTIc parameter files.

Due to the rotations performed above, this means all images produced will be ndarrays oriented in the same way. I 
am not clear on how standard ELViS data products are oriented, but it may be that we require rotations before writing
them to fits to oriented them in the way they are observed (VIS_CTI has tools for this, as I'm sure ELViS does too).
"""


def non_uniform_frame_for_ou_sim(
    ccd_id,
    quadrant_id,
    ci_normalization,
    parallel_size=2086,
    serial_size=2128,
    serial_overscan_size=29,
    pixel_scales=0.1,
):
    """
    Returns a charge injection line image suitable for OU-SIM to run through the ElVIS simulator.

    By default, this frame has dimensions (2086, 2128), representing a Euclid quadrant with a serial prescan of
    size 51 pixels, a serial overscan with 29 pixels and parallel overscan with 20 pixels.

    The charge injection line pattern is simulated using the VIS_CTI Processing element, and includes
    effects such as a non-uniform charge injection pattern. The charge injection is simulated in 3 distinct
    regions on the quadrant.

    This function assumes the same orientation for the charge injection line image, irrespective of the Euclid
    CCD ID and Quadrant ID. The orientation that it assumes has arctic clock the ndarray towards [0, 0]. However,
    based on an input CCD ID and Quadrant ID, the array is rotated to match Euclid clocking.

    Parameters
    ----------
    ccd_id : str
        The CCD ID of Euclid (runs 1 through 6)
    quadrant_id : str
        The quadrant id (E, F, G, H)
    ci_normalization : float
        The normalization of the charge injection region.
    parallel_size : int
        The size of the image in the parallel clocking direction (e.g. number of rows).
    serial_size : int
        The size of the image in the serial clocking direction (e.g. number of columns).
    serial_overscan_size : int
        The size of the serial overscan
    pixel_scales : (float, float)
        The arc-second to pixel scale conversion factor.

    Returns
    -------
    ndarray
        The charge injection line image oriented to match a given Euclid quadrant.
    """

    shape_2d = (parallel_size, serial_size)

    """Specify the charge injection regions on the CCD, which in this case is 3 equally spaced rectangular blocks."""

    ci_regions = [
        (0, 450, 51, shape_2d[1] - serial_overscan_size),
        (650, 1100, 51, shape_2d[1] - serial_overscan_size),
        (1300, 1750, 51, shape_2d[1] - serial_overscan_size),
    ]

    """Use the charge injection normalizations and regions to create *CIPatternNonUniform* of every image we'll simulate."""
    ci_pattern = pattern.CIPatternNonUniform(
        normalization=ci_normalization,
        regions=ci_regions,
        row_slope=0.0,
        column_sigma=100.0,
        maximum_normalization=84700,
    )

    """Create every pre-cti charge injection image using each *CIPattern*"""
    ci_pre_cti = ci_pattern.ci_pre_cti_from(
        shape_2d=shape_2d, pixel_scales=pixel_scales
    )

    roe_corner = euclid.roe_corner_from(ccd_id=ccd_id, quadrant_id=quadrant_id)

    """Before passing this image to arCTIc to add CTI, we want to make it a *Frame* object, which:

        - Uses an input read-out electronics corner to perform all rotations of the image before / after adding CTI.
        - Also uses this corner to rotate images before outputting to .fits, such that *Frame* objects can be load via
         .fits with the correct orientation.
        - Includes information on different regions of the image, such as the serial prescan and overscans.
    """
    return frame_util.rotate_array_from_roe_corner(
        array=ci_pre_cti, roe_corner=roe_corner
    )


def add_cti_to_frame_for_ou_sim(frames):

    """
    The *Clocker* models the CCD read-out, including CTI.

    For parallel clocking, we use 'charge injection mode' which transfers the charge of every pixel over the full CCD.
    """
    clocker = clk.Clocker(
        parallel_express=2, parallel_charge_injection_mode=True, serial_express=2
    )

    """
    The CTI model used by arCTIc to add CTI to the input image in the parallel direction, which contains: 

        - 2 *TrapInstantCapture* species in the parallel direction.
        - A simple CCD volume beta parametrization.
        - 3 *TrapInstantCapture* species in the serial direction.
        - A simple CCD volume beta parametrization.
    """
    parallel_trap_0 = traps.TrapInstantCapture(density=0.13, release_timescale=1.25)
    parallel_trap_1 = traps.TrapInstantCapture(density=0.25, release_timescale=4.4)
    parallel_ccd = ccd.CCD(
        well_fill_power=0.8, well_notch_depth=0.0, full_well_depth=84700
    )
    serial_trap_0 = traps.TrapInstantCapture(density=0.0442, release_timescale=0.8)
    serial_trap_1 = traps.TrapInstantCapture(density=0.1326, release_timescale=4.0)
    serial_trap_2 = traps.TrapInstantCapture(density=3.9782, release_timescale=20.0)
    serial_ccd = ccd.CCD(
        well_fill_power=0.8, well_notch_depth=0.0, full_well_depth=84700
    )

    """
    To simulate charge injection image, we pass the pre-cti charge injection images above through a 
    *SimulatorCIImaging*.
    
    For use outside Euclid, the Simulator includes effects like adding read-noise to the data. We will disable these
    features given that this is handled by ELViS.
    """
    simulator = ci_imaging.SimulatorCIImaging(add_poisson_noise=False)

    ci_datasets = [
        simulator.from_ci_pre_cti(
            clocker=clocker,
            ci_pre_cti=frame.ci_pre_cti,
            ci_pattern=frame.ci_pattern,
            parallel_traps=[parallel_trap_0, parallel_trap_1],
            parallel_ccd=parallel_ccd,
            serial_traps=[serial_trap_0, serial_trap_1, serial_trap_2],
            serial_ccd=serial_ccd,
        )
        for frame in zip(frames)
    ]

    """
    Finally return the images to ELViS.
    """
    return [ci_dataset.image for ci_dataset in ci_datasets]



from autocti.plot import ci_plotter_util
from autoarray.util import array_util


def plot_ci_data_for_phase(
    ci_datas_extracted,
    extract_array_from_mask,
    plot_as_subplot,
    plot_image,
    plot_noise_map,
    plot_ci_pre_cti,
    plot_signal_to_noise_map,
    plot_parallel_front_edge_line,
    plot_parallel_trails_line,
    plot_serial_front_edge_line,
    plot_serial_trails_line,
    visualize_path,
):

    plot_ci_data_arrays_for_phase(
        ci_datas_extracted=ci_datas_extracted,
        plot_as_subplot=plot_as_subplot,
        plot_image=plot_image,
        plot_noise_map=plot_noise_map,
        plot_ci_pre_cti=plot_ci_pre_cti,
        plot_signal_to_noise_map=plot_signal_to_noise_map,
        visualize_path=visualize_path,
    )

    plot_ci_data_lines_for_phase(
        ci_datas_extracted=ci_datas_extracted,
        plot_as_subplot=plot_as_subplot,
        plot_image=plot_image,
        plot_noise_map=plot_noise_map,
        plot_ci_pre_cti=plot_ci_pre_cti,
        plot_signal_to_noise_map=plot_signal_to_noise_map,
        plot_parallel_front_edge_line=plot_parallel_front_edge_line,
        plot_parallel_trails_line=plot_parallel_trails_line,
        plot_serial_front_edge_line=plot_serial_front_edge_line,
        plot_serial_trails_line=plot_serial_trails_line,
        visualize_path=visualize_path,
    )


def plot_ci_data_arrays_for_phase(
    ci_datas_extracted,
    extract_array_from_mask,
    plot_as_subplot,
    plot_image,
    plot_noise_map,
    plot_ci_pre_cti,
    plot_signal_to_noise_map,
    visualize_path,
):

    for data_index in range(len(ci_datas_extracted)):

        normalization = ci_datas_extracted[data_index].ci_pattern.normalization
        output_path = (
            visualize_path
            + "/"
            + "ci_image_"
            + str(int(normalization))
            + "/structures/"
        )
        array_util.make_path_if_does_not_exist(path=output_path + "fits/")

        if plot_as_subplot:

            plot_ci_subplot(
                ci_data=ci_datas_extracted[data_index],
                mask=ci_datas_extracted[data_index].mask,
                output_path=output_path,
                format="png",
            )

        plot_ci_data_individual(
            ci_data=ci_datas_extracted[data_index],
            mask=ci_datas_extracted[data_index].mask,
            plot_image=plot_image,
            plot_noise_map=plot_noise_map,
            plot_ci_pre_cti=plot_ci_pre_cti,
            plot_signal_to_noise_map=plot_signal_to_noise_map,
            output_path=output_path,
            format="png",
        )


def plot_ci_data_lines_for_phase(
    ci_datas_extracted,
    plot_as_subplot,
    plot_image,
    plot_noise_map,
    plot_ci_pre_cti,
    plot_signal_to_noise_map,
    plot_parallel_front_edge_line,
    plot_parallel_trails_line,
    plot_serial_front_edge_line,
    plot_serial_trails_line,
    visualize_path,
):

    line_regions = ci_plotter_util.line_regions_from_plots(
        plot_parallel_front_edge_line=plot_parallel_front_edge_line,
        plot_parallel_trails_line=plot_parallel_trails_line,
        plot_serial_front_edge_line=plot_serial_front_edge_line,
        plot_serial_trails_line=plot_serial_trails_line,
    )

    for data_index in range(len(ci_datas_extracted)):

        for line_region in line_regions:

            normalization = ci_datas_extracted[data_index].ci_pattern.normalization
            output_path = (
                visualize_path
                + "/"
                + "ci_image_"
                + str(int(normalization))
                + "/"
                + line_region
                + "/"
            )
            array_util.make_path_if_does_not_exist(path=output_path + "fits/")

            if plot_as_subplot:

                plot_ci_line_subplot(
                    ci_data=ci_datas_extracted[data_index],
                    line_region=line_region,
                    output_path=output_path,
                    format="png",
                )

            plot_ci_data_line_individual(
                ci_data=ci_datas_extracted[data_index],
                mask=ci_datas_extracted[data_index].mask,
                line_region=line_region,
                plot_image=plot_image,
                plot_noise_map=plot_noise_map,
                plot_ci_pre_cti=plot_ci_pre_cti,
                plot_signal_to_noise_map=plot_signal_to_noise_map,
                output_path=output_path,
                format="png",
            )

            plot_ci_data_line_individual(
                ci_data=ci_datas_extracted[data_index],
                mask=ci_datas_extracted[data_index].mask,
                line_region=line_region,
                plot_image=plot_image,
                plot_noise_map=plot_noise_map,
                plot_ci_pre_cti=plot_ci_pre_cti,
                plot_signal_to_noise_map=plot_signal_to_noise_map,
                output_path=output_path,
                output_format="fits",
            )



from functools import partial

import numpy as np
from autocti.charge_injection import ci_fit
from autocti.pipeline import visualizer
from autocti.pipeline.phase.dataset import analysis as analysis_dataset


class Analysis(analysis_dataset.Analysis):
    def __init__(
        self, masked_ci_imagings, clocker, settings_cti, results=None, pool=None
    ):

        super().__init__(
            masked_ci_datasets=masked_ci_imagings,
            clocker=clocker,
            settings_cti=settings_cti,
            results=results,
        )

        self.visualizers = [
            visualizer.PhaseCIImagingVisualizer(masked_dataset=masked_ci_imaging)
            for masked_ci_imaging in masked_ci_imagings
        ]

        self.pool = pool or analysis_dataset.ConsecutivePool

    @property
    def masked_ci_imagings(self):
        return self.masked_ci_datasets

    def log_likelihood_function(self, instance):
        """
        Determine the fitness of a particular model

        Parameters
        ----------
        instance

        Returns
        -------
        fit: ci_fit.Fit
            How fit the model is and the model
        """

        self.settings_cti.check_total_density_within_range(
            parallel_traps=instance.parallel_traps, serial_traps=instance.serial_traps
        )

        hyper_noise_scalars = self.hyper_noise_scalars_from_instance(instance=instance)

        pipe_cti_pass = partial(
            pipe_cti,
            instance=instance,
            clocker=self.clocker,
            hyper_noise_scalars=hyper_noise_scalars,
        )

        return np.sum(list(self.pool.map(pipe_cti_pass, self.masked_ci_datasets)))

    def hyper_noise_scalars_from_instance(self, instance, hyper_noise_scale=True):

        if not hyper_noise_scale:
            return None

        hyper_noise_scalars = list(
            filter(
                None,
                [
                    instance.hyper_noise_scalar_of_ci_regions,
                    instance.hyper_noise_scalar_of_parallel_trails,
                    instance.hyper_noise_scalar_of_serial_trails,
                    instance.hyper_noise_scalar_of_serial_overscan_no_trails,
                ],
            )
        )

        if hyper_noise_scalars:
            return hyper_noise_scalars

    def fit_from_instance_and_ci_imaging(
        self, instance, ci_imaging, hyper_noise_scale=True
    ):

        hyper_noise_scalars = self.hyper_noise_scalars_from_instance(
            instance=instance, hyper_noise_scale=hyper_noise_scale
        )

        if len(instance.parallel_traps) > 0:
            parallel_traps = list(instance.parallel_traps)
        else:
            parallel_traps = None

        if len(instance.serial_traps) > 0:
            serial_traps = list(instance.serial_traps)
        else:
            serial_traps = None

        ci_post_cti = self.clocker.add_cti(
            image=ci_imaging.ci_pre_cti,
            parallel_traps=parallel_traps,
            parallel_ccd=instance.parallel_ccd,
            serial_traps=serial_traps,
            serial_ccd=instance.serial_ccd,
        )

        return ci_fit.CIFitImaging(
            masked_ci_imaging=ci_imaging,
            ci_post_cti=ci_post_cti,
            hyper_noise_scalars=hyper_noise_scalars,
        )

    def fits_from_instance(self, instance, hyper_noise_scale=True):

        return [
            self.fit_from_instance_and_ci_imaging(
                instance=instance,
                ci_imaging=masked_ci_imaging,
                hyper_noise_scale=hyper_noise_scale,
            )
            for masked_ci_imaging in self.masked_ci_imagings
        ]

    def fits_full_dataset_from_instance(self, instance, hyper_noise_scale=True):

        return [
            self.fit_from_instance_and_ci_imaging(
                instance=instance,
                ci_imaging=masked_ci_imaging.ci_imaging_full,
                hyper_noise_scale=hyper_noise_scale,
            )
            for masked_ci_imaging in self.masked_ci_imagings
        ]

    def visualize(self, paths, instance, during_analysis):

        fits = self.fits_from_instance(instance=instance)

        for fit, visualizer in zip(fits, self.visualizers):

            visualizer.visualize_ci_imaging(paths=paths)
            visualizer.visualize_ci_fit(
                paths=paths, fit=fit, during_analysis=during_analysis
            )
        #    visualizer.visualize_ci_fit_lines(paths=paths, fit=fit, line_region="parallel_front_edge", during_analysis=during_analysis)


def pipe_cti(ci_data_masked, instance, clocker, hyper_noise_scalars):

    # TODO : Convesions ini pyarctic make this dodgy - will fix but sorting them out in arcticpy.

    if len(instance.parallel_traps) > 0:
        parallel_traps = list(instance.parallel_traps)
    else:
        parallel_traps = None

    if len(instance.serial_traps) > 0:
        serial_traps = list(instance.serial_traps)
    else:
        serial_traps = None

    ci_post_cti = clocker.add_cti(
        image=ci_data_masked.ci_pre_cti,
        parallel_traps=parallel_traps,
        parallel_ccd=instance.parallel_ccd,
        serial_traps=serial_traps,
        serial_ccd=instance.serial_ccd,
    )

    fit = ci_fit.CIFitImaging(
        masked_ci_imaging=ci_data_masked,
        ci_post_cti=ci_post_cti,
        hyper_noise_scalars=hyper_noise_scalars,
    )

    return fit.log_likelihood



from pathlib import Path

import pytest

from autoconf import conf
from autocti.mock import fixtures


directory = Path(__file__).parent.parent


@pytest.fixture(autouse=True)
def set_config_path():
    conf.instance.push(directory / "config", directory / "output")


### Arctic ###


@pytest.fixture(name="trap_0")
def make_trap_0():
    return fixtures.make_trap_0()


@pytest.fixture(name="trap_1")
def make_trap_1():
    return fixtures.make_trap_1()


@pytest.fixture(name="traps_x1")
def make_traps_x1():
    return fixtures.make_traps_x1()


@pytest.fixture(name="traps_x2")
def make_traps_x2():
    return fixtures.make_traps_x2()


@pytest.fixture(name="ccd")
def make_ccd():
    return fixtures.make_ccd()


@pytest.fixture(name="ccd_complex")
def make_ccd_complex():
    return fixtures.make_ccd_complex()


@pytest.fixture(name="parallel_clocker")
def make_parallel_clocker():
    return fixtures.make_parallel_clocker()


@pytest.fixture(name="serial_clocker")
def make_serial_clocker():
    return fixtures.make_serial_clocker()


### MASK ###


@pytest.fixture(name="mask_7x7")
def make_mask_7x7():
    return fixtures.make_mask_7x7()


### FRAMES ###


@pytest.fixture(name="scans_7x7")
def make_scans_7x7():
    return fixtures.make_scans_7x7()


@pytest.fixture(name="image_7x7")
def make_image_7x7():
    return fixtures.make_image_7x7()


@pytest.fixture(name="noise_map_7x7")
def make_noise_map_7x7():
    return fixtures.make_noise_map_7x7()


### IMAGING ###


@pytest.fixture(name="imaging_7x7")
def make_imaging_7x7():
    return fixtures.make_imaging_7x7()


### CHARGE INJECTION FRAMES ###


@pytest.fixture(name="ci_pattern_7x7")
def make_ci_pattern_7x7():
    return fixtures.make_ci_pattern_7x7()


@pytest.fixture(name="ci_image_7x7")
def make_ci_image_7x7():
    return fixtures.make_ci_image_7x7()


@pytest.fixture(name="ci_noise_map_7x7")
def make_ci_noise_map_7x7():
    return fixtures.make_ci_noise_map_7x7()


@pytest.fixture(name="ci_pre_cti_7x7")
def make_ci_pre_cti_7x7():
    return fixtures.make_ci_pre_cti_7x7()


@pytest.fixture(name="ci_cosmic_ray_map_7x7")
def make_ci_cosmic_ray_map_7x7():
    return fixtures.make_ci_cosmic_ray_map_7x7()


@pytest.fixture(name="ci_noise_scaling_maps_7x7")
def make_ci_noise_scaling_maps_7x7():

    return fixtures.make_ci_noise_scaling_maps_7x7()


### CHARGE INJECTION IMAGING ###


@pytest.fixture(name="ci_imaging_7x7")
def make_ci_imaging_7x7():

    return fixtures.make_ci_imaging_7x7()


@pytest.fixture(name="masked_ci_imaging_7x7")
def make_masked_ci_imaging_7x7():
    return fixtures.make_masked_ci_imaging_7x7()


### CHARGE INJECTION FITS ###


@pytest.fixture(name="hyper_noise_scalars")
def make_hyper_noise_scalars():
    return fixtures.make_hyper_noise_scalars()


@pytest.fixture(name="ci_fit_7x7")
def make_ci_fit_7x7():
    return fixtures.make_ci_fit_7x7()


# ### PHASES ###

from autofit.mapper.model import ModelInstance


@pytest.fixture(name="samples_with_result")
def make_samples_with_result():
    return fixtures.make_samples_with_result()


@pytest.fixture(name="phase_dataset_7x7")
def make_phase_data():
    return fixtures.make_phase_data()


@pytest.fixture(name="phase_ci_imaging_7x7")
def make_phase_ci_imaging_7x7():
    return fixtures.make_phase_ci_imaging_7x7()


# Datasets


@pytest.fixture(name="euclid_data")
def make_euclid_data():
    return fixtures.make_euclid_data()


@pytest.fixture(name="acs_ccd")
def make_acs_ccd():
    return fixtures.make_acs_ccd()


@pytest.fixture(name="acs_quadrant")
def make_acs_quadrant():
    return fixtures.make_acs_quadrant()



from autoconf import conf

from autocti.mask import mask as msk
from autocti.charge_injection import ci_mask as ci_msk
from autocti.charge_injection import ci_imaging
from autocti import exc


class SettingsCTI:
    def __init__(
        self, parallel_total_density_range=None, serial_total_density_range=None
    ):

        self.parallel_total_density_range = parallel_total_density_range
        self.serial_total_density_range = serial_total_density_range

    @property
    def tag(self):
        return (
            f"{conf.instance['notation']['settings_tags']['cti']['cti']}["
            f"{self.parallel_total_density_range_tag}"
            f"{self.serial_total_density_range_tag}]"
        )

    @property
    def parallel_total_density_range_tag(self):
        """Generate a parallel_total_density_range tag, to customize phase names based on the range of values in total \
        density that are allowed for the `NonLinearSearch` in the parallel direction.

        This changes the phase settings folder as follows:

        parallel_total_density_range = None -> settings
        parallel_total_density_range = (0, 10) -> settings__parallel_total_density_range_(0,10)
        parallel_total_density_range = (20, 60) -> settings__parallel_total_density_range_(20,60)
        """
        if self.parallel_total_density_range == None:
            return ""
        else:
            x0 = str(self.parallel_total_density_range[0])
            x1 = str(self.parallel_total_density_range[1])
            return f"__{conf.instance['notation']['settings_tags']['cti']['parallel_total_density_range']}_({x0},{x1})"

    @property
    def serial_total_density_range_tag(self):
        """Generate a serial_total_density_range tag, to customize phase names based on the range of values in total \
        density that are allowed for the `NonLinearSearch` in the serial direction.

        This changes the phase settings folder as follows:

        serial_total_density_range = None -> settings
        serial_total_density_range = (0, 10) -> settings__serial_total_density_range_(0,10)
        serial_total_density_range = (20, 60) -> settings__serial_total_density_range_(20,60)
        """
        if self.serial_total_density_range == None:
            return ""
        else:
            x0 = str(self.serial_total_density_range[0])
            x1 = str(self.serial_total_density_range[1])
            return f"__{conf.instance['notation']['settings_tags']['cti']['serial_total_density_range']}_({x0},{x1})"

    def check_total_density_within_range(self, parallel_traps, serial_traps):

        if self.parallel_total_density_range is not None:

            total_density = sum([trap.density for trap in parallel_traps])

            if (
                total_density < self.parallel_total_density_range[0]
                or total_density > self.parallel_total_density_range[1]
            ):
                raise exc.PriorException

        if self.serial_total_density_range is not None:

            total_density = sum([trap.density for trap in serial_traps])

            if (
                total_density < self.serial_total_density_range[0]
                or total_density > self.serial_total_density_range[1]
            ):
                raise exc.PriorException


class SettingsPhaseCIImaging:
    def __init__(
        self,
        settings_cti=SettingsCTI(),
        settings_mask=msk.SettingsMask(),
        settings_ci_mask=ci_msk.SettingsCIMask(),
        settings_masked_ci_imaging=ci_imaging.SettingsMaskedCIImaging(),
    ):

        self.settings_cti = settings_cti
        self.settings_mask = settings_mask
        self.settings_ci_mask = settings_ci_mask
        self.settings_masked_ci_imaging = settings_masked_ci_imaging

    @property
    def phase_tag(self):

        return (
            f"{conf.instance['notation']['settings_tags']['phase']['settings']}__"
            f"{self.settings_cti.tag}_"
            f"{self.settings_mask.tag}_"
            f"{self.settings_ci_mask.tag}_"
            f"{self.settings_masked_ci_imaging.tag}"
        )
